@@ -8,10 +8,14 @@ fi
 
 APP_HOME="$(cd "${FWDIR}/.."; pwd)"
 
-VERSION_MAJOR="1"
-VERSION_MINOR="99"
-VERSION_BUILD="20"
-VERSION=${VERSION_MAJOR}.${VERSION_MINOR}.${VERSION_BUILD}
+VERSION_MAJOR="2"
+VERSION_MINOR="1"
+VERSION_BUILD="1"
+if [ -n "$VERSION_BUILD" ]; then
+  VERSION=${VERSION_MAJOR}.${VERSION_MINOR}.${VERSION_BUILD}
+else
+  VERSION=${VERSION_MAJOR}.${VERSION_MINOR}
+fi
 
 ln -s $APP_HOME/packages/debian $APP_HOME/debian
 rm $APP_HOME/configure
