@@ -154,7 +154,12 @@ get_env_child (tree t, int i, tree env) {
 
 tree
 get_env_descendant (tree t, path p, tree env) {
-  return the_drd->get_env_descendant (t, p, env);  
+  return the_drd->get_env_descendant (t, p, env);
+}
+
+tree
+get_env_descendant (tree t, path p, string var, tree val) {
+  return the_drd->get_env_descendant (t, p, var, val);
 }
 
 /******************************************************************************
@@ -166,19 +171,27 @@ move_any (tree t, path p, bool forward) {
   path q = path_up (p);
   int  l = last_item (p);
   tree st= subtree (t, q);
+  if (!is_nil (q) && is_func (subtree (t, path_up (q)), RAW_DATA)) {
+    if (forward) return path_up (q) * 1;
+    else return path_up (q) * 0;
+  }
   if (is_atomic (st)) {
     string s= st->label;
+#ifdef SANITY_CHECKS
     ASSERT (l >= 0 && l <= N(s), "out of range");
+#else
+    l= max (min (l, N(s)), 0);
+#endif
     if (forward) {
       if (l<N(s)) {
-	tm_char_forwards (s, l);
-	return q * l;
+        tm_char_forwards (s, l);
+        return q * l;
       }
     }
     else {
       if (l>0) {
-	tm_char_backwards (s, l);
-	return q * l;
+        tm_char_backwards (s, l);
+        return q * l;
       }
     }
   }
@@ -186,13 +199,13 @@ move_any (tree t, path p, bool forward) {
     int i, n= N(st);
     if (forward) {
       for (i=0; i<n; i++)
-	if (the_drd->is_accessible_child (st, i))
-	  return q * path (i, 0);
+        if (the_drd->is_accessible_child (st, i))
+          return q * path (i, 0);
     }
     else {
       for (i=n-1; i>=0; i--)
-	if (the_drd->is_accessible_child (st, i))
-	  return q * path (i, right_index (st[i]));
+        if (the_drd->is_accessible_child (st, i))
+          return q * path (i, right_index (st[i]));
     }
     return q * (1-l);
   }
@@ -205,12 +218,12 @@ move_any (tree t, path p, bool forward) {
   if (forward) {
     for (i=l+1; i<n; i++)
       if (the_drd->is_accessible_child (st, i))
-	return q * path (i, 0);
+        return q * path (i, 0);
   }
   else {
     for (i=l-1; i>=0; i--)
       if (the_drd->is_accessible_child (st, i)) {
-	return q * path (i, right_index (st[i]));
+        return q * path (i, right_index (st[i]));
     }
   }
   return q * (forward? 1: 0);
@@ -225,9 +238,36 @@ path previous_any (tree t, path p) {
 * Traversal of a valid cursor positions inside a tree
 ******************************************************************************/
 
+path
+closest_up (tree t, path p) {
+  if (is_atomic (t)) {
+    if (is_nil (p)) return path (0);
+    else return path (max (0, min (p->item, N(t->label))));
+  }
+  else if (is_concat (t) || is_document (t)) {
+    if (N(t) == 0) return path (0);
+    else if (is_nil (p) || is_atom (p) || p->item < 0)
+      return path (0, closest_up (t[0], path ()));
+    else if (p->item >= N(t))
+      return path (N(t)-1, closest_up (t[N(t)-1], path ()));
+    else return path (p->item, closest_up (t[p->item], p->next));
+  }
+  else {
+    if (is_nil (p)) return path (0);
+    else if (is_atom (p)) return path (max (0, min (p->item, 1)));
+    else if (p->item < 0) return path (0);
+    else if (p->item >= N(t)) return path (1);
+    else return path (p->item, closest_up (t[p->item], p->next));
+  }
+}
+
 static path
 move_valid_sub (tree t, path p, bool forward) {
-  ASSERT (is_inside (t, p), "invalid cursor");
+#ifdef SANITY_CHECKS
+  ASSERT (is_inside (t, p), "invalid cursor [move_valid]");
+#else
+  if (!is_inside (t, p)) p= closest_up (t, p);
+#endif
   path q= p;
   while (true) {
     path r= move_any (t, q, forward);
@@ -238,7 +278,7 @@ move_valid_sub (tree t, path p, bool forward) {
 }
 
 static path
-move_valid (tree t, path p, bool forward) {
+move_valid_bis (tree t, path p, bool forward) {
   bool inside= the_drd->is_accessible_path (t, p);
   if (inside) return move_valid_sub (t, p, forward);
   bool old_mode= set_access_mode (DRD_ACCESS_SOURCE);
@@ -247,6 +287,22 @@ move_valid (tree t, path p, bool forward) {
   return r;
 }
 
+static bool
+inside_graphics (tree t, path p) {
+  if (is_func (t, GRAPHICS)) return true;
+  if (is_nil (p) || is_nil (p->next)) return false;
+  return inside_graphics (t[p->item], p->next);
+}
+
+static path
+move_valid (tree t, path p, bool forward) {
+  // NOTE: extra hook for moving inside graphical text
+  path q= move_valid_bis (t, p, forward);
+  if (!inside_graphics (t, p)) return q;
+  if (inside_contiguous_document (t, p, q)) return q;
+  return false;
+}
+
 path next_valid (tree t, path p) {
   return move_valid (t, p, true); }
 path previous_valid (tree t, path p) {
@@ -254,7 +310,11 @@ path previous_valid (tree t, path p) {
 
 static path
 move_accessible (tree t, path p, bool forward) {
-  ASSERT (is_inside (t, p), "invalid cursor");
+#ifdef SANITY_CHECKS
+  ASSERT (is_inside (t, p), "invalid cursor [move_accessible]");
+#else
+  if (!is_inside (t, p)) p= closest_up (t, p);
+#endif
   path q= p;
   while (true) {
     path r= move_any (t, q, forward);
@@ -299,7 +359,7 @@ next_is_word (tree t, path p) {
 static path
 move_word (tree t, path p, bool forward) {
   while (true) {
-    path q= move_valid (t, p, forward);
+    path q= move_accessible (t, p, forward);
     int l= last_item (q);
     if (q == p) return p;
     tree st= subtree (t, path_up (q));
@@ -308,23 +368,23 @@ move_word (tree t, path p, bool forward) {
       int n= N(s);
       if (s == "") return q;
       if (forward && l>0 &&
-	  (is_iso_alphanum (s[l-1]) ||
-	   (l==n && at_border (t, path_up (q), forward))) &&
-	  (l==n || !is_iso_alphanum (s[l])))
-	return q;
+          (is_iso_alphanum (s[l-1]) || (l<n && s[l] == ' ') ||
+           (l==n && at_border (t, path_up (q), forward))) &&
+          (l==n || !is_iso_alphanum (s[l])))
+        return q;
       if (!forward && l<n &&
-	  (is_iso_alphanum (s[l]) ||
-	   (l==0 && at_border (t, path_up (q), forward))) &&
-	  (l==0 || !is_iso_alphanum (s[l-1])))
-	return q;
+          (is_iso_alphanum (s[l]) || (l>0 && s[l-1] == ' ') ||
+           (l==0 && at_border (t, path_up (q), forward))) &&
+          (l==0 || !is_iso_alphanum (s[l-1])))
+        return q;
       if (!forward && l==n && next_is_word (t, path_up (q)))
-	return q;
+        return q;
     }
     else {
       if (forward && l==1) return q;
       if (!forward && l==0) return q;
       if (!forward && next_is_word (t, path_up (q)))
-	return q;
+        return q;
     }
     p= q;
   }
@@ -346,7 +406,7 @@ move_node (tree t, path p, bool forward) {
     if (forward) p= path_up (p) * N (st->label);
     else p= path_up (p) * 0;
   }
-  p= move_valid (t, p, forward);
+  p= move_accessible (t, p, forward);
   st= subtree (t, path_up (p));
   if (is_atomic (st) && last_item (p) > 0)
     p= path_up (p) * N (st->label);
@@ -380,7 +440,7 @@ distinct_tag_or_argument (tree t, path p, path q, hashset<int> labs) {
   while (!is_nil (r) && (r != c)) {
     r= path_up (r);
     if (labs->contains ((int) L (subtree (t, r))) &&
-	!none_accessible (subtree (t, r)))
+        !none_accessible (subtree (t, r)))
       return true;
   }
   return false;
@@ -394,6 +454,23 @@ acceptable_border (tree t, path p, path q, hashset<int> labs) {
   return tag_border (t, p) > 0;
 }
 
+static bool
+acceptable_child (tree t, path p, hashset<int> labs) {
+  tree st= subtree (t, path_up (p));
+  if (is_compound (st) && labs->contains ((int) L(st)))
+    if (last_item (p) == 0 || last_item (p) == 1)
+      return true;
+  p= path_up (p);
+  while (!is_nil (p)) {
+    st= subtree (t, path_up (p));
+    if (labs->contains ((int) L(st)))
+      if (is_accessible_child (st, last_item (p)))
+        return true;
+    p= path_up (p);
+  }
+  return false;
+}
+
 static int
 tag_index (tree t, path p, hashset<int> labs) {
   p= path_up (p);
@@ -405,15 +482,25 @@ tag_index (tree t, path p, hashset<int> labs) {
   return -1;
 }
 
+static bool
+tag_inside (tree t, path p, hashset<int> labs) {
+  return !labs->contains ((int) L (subtree (t, path_up (p))));
+}
+
 static path
 move_tag (tree t, path p, hashset<int> labs, bool forward, bool preserve) {
   path q= p;
   while (true) {
     path r= move_node (t, q, forward);
     if (r == q) return p;
+    if (!tag_inside (t, r, labs) && !tag_inside (t, p, labs) &&
+        last_item (r) == last_item (p))
+      return r;
     if (distinct_tag_or_argument (t, p, r, labs) &&
         acceptable_border (t, p, r, labs) &&
-	(!preserve || tag_index (t, r, labs) == tag_index (t, p, labs)))
+        acceptable_child (t, r, labs) &&
+        tag_inside (t, r, labs) == tag_inside (t, p, labs) &&
+        (!preserve || tag_index (t, r, labs) == tag_index (t, p, labs)))
       return r;
     q= r;
   }
@@ -428,7 +515,7 @@ get_labels (scheme_tree t) {
     int i, n= N(t);
     for (i=0; i<n; i++)
       if (is_atomic (t[i]))
-	labs->insert ((int) as_tree_label (t[i]->label));
+        labs->insert ((int) as_tree_label (t[i]->label));
   }
   return labs;
 }
@@ -456,12 +543,12 @@ move_argument (tree t, path p, bool forward) {
   if (forward) {
     for (i=l+1; i<n; i++)
       if (the_drd->is_accessible_child (st, i))
-	return start (t, q * i);
+        return start (t, q * i);
   }
   else {
     for (i=l-1; i>=0; i--)
       if (the_drd->is_accessible_child (st, i))
-	return end (t, q * i);
+        return end (t, q * i);
   }
   return path ();
 }
@@ -475,24 +562,75 @@ path previous_argument (tree t, path p) {
 * Other routines
 ******************************************************************************/
 
-static path
-search_upwards (tree t, path p, tree_label which) {
-  if (is_nil (p) || L (subtree (t, p)) == which) return p;
-  else return search_upwards (t, path_up (p), which);
+path
+search_common (tree t, path p, path q, tree_label which) {
+  if (is_atomic (t)) return path ();
+  if (is_nil (p) || is_nil (q) || p->item != q->item) return path ();
+  if (0 > p->item || p->item >= N(t)) return path ();
+  path c= search_common (t[p->item], p->next, q->next, which);
+  if (L(subtree (t[p->item], c)) == which) return path (p->item, c);
+  return path ();
+}
+
+bool
+no_other (tree t, path p, tree_label which) {
+  if (L(t) == which) return false;
+  if (is_nil (p)) return true;
+  if (0 > p->item || p->item >= N(t)) return false;
+  return no_other (t[p->item], p->next, which);
 }
 
 bool
 inside_same (tree t, path p, path q, tree_label which) {
-  return
-    search_upwards (t, path_up (p), which) ==
-    search_upwards (t, path_up (q), which);
+  path d= search_common (t, path_up (p), path_up (q), which);
+  if (L(subtree (t, d)) != which) return true;
+  path sp= p / d;
+  path sq= q / d;
+  if (is_atom (sp) || is_atom (sq)) return false;
+  if (sp->item != sq->item) return false;
+  if (0 > sp->item || sp->item >= N(subtree (t, d))) return false;
+  tree st= subtree (t, d) [sp->item];
+  return no_other (st, path_up (sq->next), which) ||
+         no_other (st, path_up (sp->next), which);
 }
 
 bool
-more_inside (tree t, path p, path q, tree_label which) {
-  return
-    search_upwards (t, path_up (q), which) <=
-    search_upwards (t, path_up (p), which);
+inside_same_or_more (tree t, path p, path q, tree_label which) {
+  path d= search_common (t, path_up (p), path_up (q), which);
+  if (L(subtree (t, d)) != which) return true;
+  path sp= p / d;
+  path sq= q / d;
+  if (is_atom (sp) || is_atom (sq)) return false;
+  if (sp->item != sq->item) return false;
+  if (0 > sp->item || sp->item >= N(subtree (t, d))) return false;
+  tree st= subtree (t, d) [sp->item];
+  return no_other (st, path_up (sq->next), which);
+}
+
+bool
+is_boundary (tree t, path p) {
+  if (is_func (subtree (t, p), DOCUMENT)) return true;
+  if (is_nil (p)) return false;
+  path q= path_up (p);
+  if (is_func (subtree (t, q), GRAPHICS)) return true;
+  if (is_nil (q)) return false;
+  if (is_func (subtree (t, path_up (q)), GRAPHICS)) return true;
+  return false;
+}
+
+bool
+inside_contiguous_document (tree t, path op, path oq) {
+  if (!inside_same (t, op, oq, DOCUMENT)) return false;
+  path p= path_up (op), q= path_up (oq);
+  while (!is_nil (p) && !is_boundary (t, p)) p= path_up (p);
+  while (!is_nil (q) && !is_boundary (t, q)) q= path_up (q);
+  if (p == q) return true;
+  if (q <= p) return inside_contiguous_document (t, oq, op);
+  if (!(p <= q)) return false;
+  if (path_less (op, oq))
+    return (oq / q) == start (subtree (t, q));
+  else
+    return (oq / q) == end (subtree (t, q));
 }
 
 /******************************************************************************
@@ -508,9 +646,10 @@ init_sections () {
     section_traverse_tags= hashset<tree_label> ();
     section_traverse_tags->insert (DOCUMENT);
     section_traverse_tags->insert (CONCAT);
-    section_traverse_tags->insert (as_tree_label ("ignore"));
-    section_traverse_tags->insert (as_tree_label ("show-part"));
-    section_traverse_tags->insert (as_tree_label ("hide-part"));
+    section_traverse_tags->insert (make_tree_label ("ignore"));
+    section_traverse_tags->insert (make_tree_label ("show-part"));
+    section_traverse_tags->insert (make_tree_label ("hide-part"));
+    section_traverse_tags->insert (make_tree_label ("live-io*"));
   }
   if (N(section_tags) == 0) {
     eval ("(use-modules (text text-drd))");
@@ -527,16 +666,18 @@ previous_section_impl (tree t, path p) {
   if (is_atomic (t)) return path ();
   else if (N(t) == 1 && section_tags->contains (L(t)))
     return p * 0;
+  else if (is_compound (t, "live-io*") && !is_nil (p) && p->item == 2)
+    return path (2, previous_section_impl (t[2], p->next));
   else if (section_traverse_tags->contains (L(t))) {
     int i= is_nil (p)? N(t)-1: p->item;
     for (; i>=0; i--) {
       if (!is_nil (p) && i == p->item) {
-	path r= previous_section_impl (t[i], p->next);
-	if (!is_nil (r)) return path (i, r);
+        path r= previous_section_impl (t[i], p->next);
+        if (!is_nil (r)) return path (i, r);
       }
       else {
-	path r= previous_section_impl (t[i], path ());
-	if (!is_nil (r)) return path (i, r);
+        path r= previous_section_impl (t[i], path ());
+        if (!is_nil (r)) return path (i, r);
       }
     }
   }
