@@ -22,18 +22,19 @@ enum tree_label {
 
   // basic formatting tags
   DOCUMENT, PARA, SURROUND, CONCAT,
-  RIGID, HIDDEN, FREEZE, UNFREEZE,
+  RIGID, HGROUP, HIDDEN, FREEZE, UNFREEZE,
   HSPACE, VAR_VSPACE, VSPACE, SPACE, HTAB,
   MOVE, SHIFT, RESIZE, CLIPPED,
-  REPEAT, _FLOAT, DATOMS, DLINES, DPAGES, DBOX,
+  REPEAT, VAR_REPEAT, _FLOAT, DATOMS, DLINES, DPAGES, DBOX,
+  LINE_NOTE, PAGE_NOTE, IF_PAGE_BREAK,
 
   // zero-ary formatting directives
   // modify is_formatting predicate when inserting new tags
   WITH_LIMITS, LINE_BREAK, NEW_LINE, NEXT_LINE,
-  NO_BREAK, YES_INDENT, NO_INDENT, VAR_YES_INDENT,
-  VAR_NO_INDENT, VAR_PAGE_BREAK, PAGE_BREAK,
-  VAR_NO_PAGE_BREAK, NO_PAGE_BREAK, VAR_NEW_PAGE,
-  NEW_PAGE, VAR_NEW_DPAGE, NEW_DPAGE,
+  NO_BREAK, YES_INDENT, NO_INDENT, VAR_YES_INDENT, VAR_NO_INDENT,
+  VAR_PAGE_BREAK, PAGE_BREAK, VAR_NO_PAGE_BREAK, NO_PAGE_BREAK,
+  VAR_NO_BREAK_HERE, NO_BREAK_HERE, NO_BREAK_START, NO_BREAK_END,
+  VAR_NEW_PAGE, NEW_PAGE, VAR_NEW_DPAGE, NEW_DPAGE,
 
   // mathematics
   AROUND, VAR_AROUND, BIG_AROUND,
@@ -48,21 +49,25 @@ enum tree_label {
   TABLE, ROW, CELL, SUBTABLE,
 
   // macro language
-  ASSIGN, WITH, PROVIDES, VALUE, QUOTE_VALUE,
+  ASSIGN, PROVIDE, WITH, PROVIDES, VALUE, QUOTE_VALUE, OR_VALUE,
   MACRO, DRD_PROPS, ARG, QUOTE_ARG, COMPOUND,
   XMACRO, GET_LABEL, GET_ARITY, MAP_ARGS, EVAL_ARGS,
+  NEW_THEME, COPY_THEME, APPLY_THEME, SELECT_THEME,
   MARK, EXPAND_AS,
   EVAL, QUOTE, QUASI, QUASIQUOTE, UNQUOTE, VAR_UNQUOTE, COPY,
   IF, VAR_IF, CASE, WHILE, FOR_EACH,
-  EXTERN, INCLUDE, USE_PACKAGE, USE_MODULE,
+  EXTERN, VAR_INCLUDE, INCLUDE,
+  WITH_PACKAGE, USE_PACKAGE, USE_MODULE,
 
   // computational markup
   OR, XOR, AND, NOT,
   PLUS, MINUS, TIMES, OVER, DIV, MOD, MINIMUM, MAXIMUM,
   MATH_SQRT, EXP, LOG, POW, COS, SIN, TAN,
-  MERGE, LENGTH, RANGE, NUMBER, _DATE, TRANSLATE, CHANGE_CASE, FIND_FILE,
-  IS_TUPLE, LOOK_UP,
+  MERGE, LENGTH, RANGE, NUMBER, _DATE, TRANSLATE, CHANGE_CASE,
+  FIND_FILE, FIND_FILE_UPWARDS,
+  IS_TUPLE, LOOK_UP, OCCURS_INSIDE,
   EQUAL, UNEQUAL, LESS, LESSEQ, GREATER, GREATEREQ,
+  BLEND, RGB_COLOR, RGB_ACCESS,
 
   // built-in length units
   CM_LENGTH, MM_LENGTH, IN_LENGTH, PT_LENGTH,
@@ -72,12 +77,16 @@ enum tree_label {
   FN_LENGTH, FNS_LENGTH, BLS_LENGTH,
   FNBOT_LENGTH, FNTOP_LENGTH,
   SPC_LENGTH, XSPC_LENGTH,                        // font specific & stretchy
-  PAR_LENGTH, PAG_LENGTH,                         // other
+  PAR_LENGTH, PAW_LENGTH, PAG_LENGTH,             // page dimensions
   GW_LENGTH, GH_LENGTH, GU_LENGTH,                // extents & unit of graphics
-  TMPT_LENGTH, PX_LENGTH,                         // miscellaneous
-  MSEC_LENGTH, SEC_LENGTH, MIN_LENGTH, HR_LENGTH,  // durations
+  TMPT_LENGTH, PX_LENGTH, GUIPX_LENGTH,           // miscellaneous
+  LCORNER_LENGTH, BCORNER_LENGTH,
+  RCORNER_LENGTH, TCORNER_LENGTH,
+  MS_LENGTH, S_LENGTH, MSEC_LENGTH, SEC_LENGTH,   // durations
+  MIN_LENGTH, HR_LENGTH,
 
   // tags for source tree editing
+  FILTER_STYLE,
   STYLE_WITH, VAR_STYLE_WITH, STYLE_ONLY, VAR_STYLE_ONLY,
   ACTIVE, VAR_ACTIVE, INACTIVE, VAR_INACTIVE,
   REWRITE_INACTIVE, INLINE_TAG, OPEN_TAG, MIDDLE_TAG, CLOSE_TAG,
@@ -85,23 +94,30 @@ enum tree_label {
 
   // linking primitives
   LOCUS, ID, HARD_ID,
-  LINK, URL, SCRIPT, HLINK, ACTION,
-  SET_BINDING, GET_BINDING, LABEL, REFERENCE, PAGEREF, WRITE,
+  LINK, URL, SCRIPT, OBSERVER,
+  FIND_ACCESSIBLE, HLINK, ACTION,
+  SET_BINDING, GET_BINDING, HAS_BINDING, HIDDEN_BINDING,
+  LABEL, REFERENCE, PAGEREF,
+  GET_ATTACHMENT, WRITE, TOC_NOTIFY,
 
   // various other tags
   TUPLE, ATTR, TMLEN, COLLECTION, ASSOCIATE, BACKUP, PATTERN, GRADIENT,
-  SPECIFIC, FLAG,
+  SPECIFIC, FLAG, HYPHENATE_AS,
 
   // animations
-  ANIM_COMPOSE, ANIM_REPEAT, ANIM_CONSTANT,
+  ANIM_STATIC, ANIM_DYNAMIC, MORPH,
+  ANIM_TIME, ANIM_PORTION,
+  ANIM_COMPOSE, ANIM_REPEAT, ANIM_CONSTANT, ANIM_ACCELERATE,
   ANIM_TRANSLATE, ANIM_PROGRESSIVE, VIDEO, SOUND,
 
   // graphical tags
   GRAPHICS,
   SUPERPOSE,
   GR_GROUP, GR_TRANSFORM, GR_EFFECT,
-  TEXT_AT, MATH_AT, _POINT,
-  LINE, CLINE, ARC, CARC, SPLINE, VAR_SPLINE, CSPLINE,
+  TEXT_AT, MATH_AT, DOCUMENT_AT, _POINT,
+  LINE, CLINE, ARC, CARC,
+  SPLINE, VAR_SPLINE, CSPLINE,
+  BEZIER, CBEZIER, SMOOTH, CSMOOTH,
   FILL, IMAGE,
   BOX_INFO, FRAME_DIRECT, FRAME_INVERSE,
   IS_EQUAL, IS_INTERSECTION,
@@ -110,15 +126,18 @@ enum tree_label {
   LIGHT_3D, LIGHT_DIFFUSE, LIGHT_SPECULAR,
 
   // graphical effects
-  EFF_MOVE, EFF_MAGNIFY, EFF_BUBBLE,
+  EFF_MOVE, EFF_MAGNIFY, EFF_BUBBLE, EFF_CROP,
+  EFF_TURBULENCE, EFF_FRACTAL_NOISE,
+  EFF_HATCH, EFF_DOTS,
   EFF_GAUSSIAN, EFF_OVAL, EFF_RECTANGULAR, EFF_MOTION,
   EFF_BLUR, EFF_OUTLINE, EFF_THICKEN, EFF_ERODE,
+  EFF_DEGRADE, EFF_DISTORT, EFF_GNAW,
   EFF_SUPERPOSE, EFF_ADD, EFF_SUB, EFF_MUL, EFF_MIN, EFF_MAX, EFF_MIX,
-  EFF_NORMALIZE, EFF_MONOCHROME, EFF_COLOR_MATRIX,
-  EFF_MAKE_TRANSPARENT, EFF_MAKE_OPAQUE,
+  EFF_NORMALIZE, EFF_MONOCHROME, EFF_COLOR_MATRIX, EFF_GRADIENT,
+  EFF_MAKE_TRANSPARENT, EFF_MAKE_OPAQUE, EFF_RECOLOR, EFF_SKIN,
 
   // graphical interface markup
-  CANVAS, ORNAMENT,
+  CANVAS, ORNAMENT, ART_BOX,
 
   // obsolete tags
   FORMAT, LINE_SEP, SPLIT, DELAY, HOLD, RELEASE,
@@ -144,11 +163,5 @@ string as_string (tree_label l);
 tree_label as_tree_label (string s);
 bool existing_tree_label (string s);
 
-#ifdef OS_WIN32
-inline bool operator == (tree_label first, tree_label second) {
-  return (int) first == (int) second; }
-inline bool operator != (tree_label first, tree_label second) {
-  return (int) first != (int) second; }
-#endif
 
 #endif // defined TREE_LABEL_H
