@@ -18,10 +18,39 @@
 #define TEX_ANY   0
 #define TEX_EC    1
 #define TEX_LA    2
-#define TEX_CM    3
-#define TEX_ADOBE 4
+#define TEX_GR    3
+#define TEX_CM    4
+#define TEX_ADOBE 5
 
 static void special_initialize ();
+font_metric tfm_font_metric (tex_font_metric tfm, font_glyphs pk, double unit);
+
+hashmap<string,double> lsub_ecrm_table ();
+hashmap<string,double> lsup_ecrm_table ();
+hashmap<string,double> rsub_ecrm_table ();
+hashmap<string,double> rsup_ecrm_table ();
+hashmap<string,double> lsub_cmr_table ();
+hashmap<string,double> lsup_cmr_table ();
+hashmap<string,double> rsub_cmr_table ();
+hashmap<string,double> rsup_cmr_table ();
+hashmap<string,double> above_cmr_table ();
+hashmap<string,double> lsub_cmmi_table ();
+hashmap<string,double> lsup_cmmi_table ();
+hashmap<string,double> rsub_cmmi_table ();
+hashmap<string,double> rsup_cmmi_table ();
+hashmap<string,double> above_cmmi_table ();
+hashmap<string,double> lsub_cmsy_table ();
+hashmap<string,double> lsup_cmsy_table ();
+hashmap<string,double> above_cmsy_table ();
+hashmap<string,double> lsub_bbm_table ();
+hashmap<string,double> lsup_bbm_table ();
+hashmap<string,double> rsub_bbm_table ();
+hashmap<string,double> rsup_bbm_table ();
+hashmap<string,double> above_bbm_table ();
+hashmap<string,double> above_eufm_table ();
+hashmap<string,double> rsub_rsfs_table ();
+hashmap<string,double> rsup_rsfs_table ();
+hashmap<string,double> above_rsfs_table ();
 
 /******************************************************************************
 * TeX text fonts
@@ -40,25 +69,34 @@ struct tex_font_rep: font_rep {
   tex_font_rep (string name, int status,
 		string family, int size, int dpi, int dsize);
 
-  bool raw_supports (unsigned char c);
-  bool supports (string c);
-  void get_extents (string s, metric& ex);
-  void get_xpositions (string s, SI* xpos);
-  void draw_fixed (renderer ren, string s, SI x, SI y);
-  font magnify (double zoom);
-  SI   get_left_correction (string s);
-  SI   get_right_correction (string s);
+  bool  raw_supports (unsigned char c);
+  bool  supports (string c);
+  void  get_extents (string s, metric& ex);
+  void  get_xpositions (string s, SI* xpos, bool ligf);
+  void  get_xpositions (string s, SI* xpos);
+  void  draw_fixed (renderer ren, string s, SI x, SI y);
+  font  magnify (double zoomx, double zoomy);
+  SI    get_left_correction  (string s);
+  SI    get_right_correction (string s);
+  SI    get_lsub_correction  (string s);
+  SI    get_lsup_correction  (string s);
+  SI    get_rsub_correction  (string s);
+  SI    get_rsup_correction  (string s);
+  SI    get_wide_correction  (string s, int mode);
+  void  advance_glyph (string s, int& pos, bool ligf);
   glyph get_glyph (string s);
-  void special_get_extents (string s, metric& ex);
-  void special_get_xpositions (string s, SI* xpos);
-  void special_draw (renderer ren, string s, SI x, SI y);
-  SI   special_get_left_correction (string s);
-  SI   special_get_right_correction (string s);
-  void accented_get_extents (string s, metric& ex);
-  void accented_get_xpositions (string s, SI* xpos);
-  void accented_draw (renderer ren, string s, SI x, SI y);
-  SI   accented_get_left_correction (string s);
-  SI   accented_get_right_correction (string s);
+  int   index_glyph (string s, font_metric& fnm, font_glyphs& fng);
+  int   get_ligature_code (string s);
+  void  special_get_extents (string s, metric& ex);
+  void  special_get_xpositions (string s, SI* xpos, bool ligf);
+  void  special_draw (renderer ren, string s, SI x, SI y);
+  SI    special_get_left_correction (string s);
+  SI    special_get_right_correction (string s);
+  void  accented_get_extents (string s, metric& ex);
+  void  accented_get_xpositions (string s, SI* xpos, bool ligf);
+  void  accented_draw (renderer ren, string s, SI x, SI y);
+  SI    accented_get_left_correction (string s);
+  SI    accented_get_right_correction (string s);
 };
 
 /******************************************************************************
@@ -87,8 +125,9 @@ tex_font_rep::tex_font_rep (string name, int status2,
   extra        = conv (tfm->spc_extra ());
   extra->min   = extra->min >> 1;
   extra->max   = extra->min << 1;
+  mspc         = spc;
   sep          = ((((dpi*PIXEL)/72)*design_size) >> 8) / 10;
-  exec         = ! ends (family, "tt");
+  exec         = !ends (family, "tt");
 
   y1           = conv (-262080);   // -0.25 quad
   y2           = y1+ display_size; //  0.75 quad
@@ -102,6 +141,7 @@ tex_font_rep::tex_font_rep (string name, int status2,
   yshift       = yx/6;
 
   wpt          = (dpi*PIXEL)/72;
+  hpt          = (dpi*PIXEL)/72;
   wfn          = (wpt*design_size) >> 8;
   wline        = wfn/20;
   wquad        = conv (tfm->spc_quad ());
@@ -118,6 +158,71 @@ tex_font_rep::tex_font_rep (string name, int status2,
   }
 
   special_initialize ();
+
+  if (family == "ecrm" || family == "ecbx") {
+    lsub_correct= lsub_ecrm_table ();
+    lsup_correct= lsup_ecrm_table ();
+    rsub_correct= rsub_ecrm_table ();
+    rsup_correct= rsup_ecrm_table ();
+    above_correct= hashmap<string,double> (0.0);
+  }
+  else if (family == "cmr" || family == "cmbx") {
+    lsub_correct= lsub_cmr_table ();
+    lsup_correct= lsup_cmr_table ();
+    rsub_correct= rsub_cmr_table ();
+    rsup_correct= rsup_cmr_table ();
+    above_correct= above_cmr_table ();
+  }
+  else if (family == "cmmi" || family == "cmmib") {
+    lsub_correct= lsub_cmmi_table ();
+    lsup_correct= lsup_cmmi_table ();
+    rsub_correct= rsub_cmmi_table ();
+    rsup_correct= rsup_cmmi_table ();
+    above_correct= above_cmmi_table ();
+  }
+  else if (family == "cmsy" || family == "cmbsy") {
+    lsub_correct= lsub_cmsy_table ();
+    lsup_correct= lsup_cmsy_table ();
+    rsub_correct= hashmap<string,double> (0.0);
+    rsup_correct= hashmap<string,double> (0.0);
+    above_correct= above_cmsy_table ();
+  }
+  else if (family == "bbm" || family == "bbmbx") {
+    lsub_correct= lsub_bbm_table ();
+    lsup_correct= lsup_bbm_table ();
+    rsub_correct= rsub_bbm_table ();
+    rsup_correct= rsup_bbm_table ();
+    above_correct= above_bbm_table ();
+  }
+  else if (family == "eufm" || family == "eufb") {
+    lsub_correct= hashmap<string,double> (0.0);
+    lsup_correct= hashmap<string,double> (0.0);
+    rsub_correct= hashmap<string,double> (0.0);
+    rsup_correct= hashmap<string,double> (0.0);
+    above_correct= above_eufm_table ();
+  }
+  else if (family == "rsfs") {
+    yx= (SI) round (0.53 * display_size);
+    yfrac        = yx >> 1;
+    ysub_lo_base = -yx/3;
+    ysub_hi_lim  = (5*yx)/6;
+    ysup_lo_lim  = yx/2;
+    ysup_lo_base = (5*yx)/6;
+    ysup_hi_lim  = yx;
+    yshift       = yx/6;
+    lsub_correct= hashmap<string,double> (0.0);
+    lsup_correct= hashmap<string,double> (0.0);
+    rsub_correct= rsub_rsfs_table ();
+    rsup_correct= rsup_rsfs_table ();
+    above_correct= above_rsfs_table ();
+  }
+  else {
+    lsub_correct= hashmap<string,double> (0.0);
+    lsup_correct= hashmap<string,double> (0.0);
+    rsub_correct= hashmap<string,double> (0.0);
+    rsup_correct= hashmap<string,double> (0.0);
+    above_correct= hashmap<string,double> (0.0);
+  }
 }
 
 /******************************************************************************
@@ -128,22 +233,33 @@ static bool special_initialized= false;
 static hashmap<string,string> special_translate ("");
 
 static void
-special_initialize () {
-  if (special_initialized) return;
-  special_translate ("<less>")= "<";
-  special_translate ("<gtr>")= ">";
-  translator trl= load_translator ("larm");
+special_initialize (string enc) {
+  translator trl= load_translator (enc);
   iterator<string> it= iterate (trl->dict);
   while (it->busy ()) {
     string s= it->next ();
     special_translate (s)= string ((char) (unsigned char) trl->dict[s]);
+    if (N(s) > 2 && s(0,2) == "<#") {
+      string sl= locase_all (s), su= upcase_all (s);
+      special_translate (sl)= string ((char) (unsigned char) trl->dict[s]);
+      special_translate (su)= string ((char) (unsigned char) trl->dict[s]);
+    }
   }
+}
+
+static void
+special_initialize () {
+  if (special_initialized) return;
+  special_translate ("<less>")= "<";
+  special_translate ("<gtr>")= ">";
+  special_initialize ("larm");
+  special_initialize ("grmn");
   special_initialized= true;
 }
 
 void
 tex_font_rep::special_get_extents (string s, metric& ex) {
-  register int i, j;
+  int i, j;
   for (i=0; i<N(s); i++)
     if (s[i]=='<') break;
   get_extents (s (0, i), ex);
@@ -175,15 +291,15 @@ tex_font_rep::special_get_extents (string s, metric& ex) {
 }
 
 void
-tex_font_rep::special_get_xpositions (string s, SI* xpos) {
+tex_font_rep::special_get_xpositions (string s, SI* xpos, bool ligf) {
   SI offset= 0;
-  register int l=0, i, j, n=N(s);
+  int l=0, i, j, n=N(s);
   while (l<n) {
     for (i=l; i<n; i++)
       if (s[i]=='<') break;
     if (l<i) {
-      get_xpositions (s (l, i), xpos + l);
-      for (j=l+1; j<=i; j++) xpos[j] += offset;
+      get_xpositions (s (l, i), xpos + l, ligf);
+      for (j=l; j<=i; j++) xpos[j] += offset;
       if (i==n) break;
       offset= xpos[i];
     }
@@ -209,7 +325,7 @@ tex_font_rep::special_get_xpositions (string s, SI* xpos) {
 
 void
 tex_font_rep::special_draw (renderer ren, string s, SI x, SI y) {
-  register int i, j;
+  int i, j;
   metric ex;
   for (i=0; i<N(s); i++)
     if (s[i]=='<') break;
@@ -377,7 +493,7 @@ tex_font_rep::accented_get_extents (string s, metric& ex) {
   int old_status= status;
   status= TEX_ANY;
 
-  register int i;
+  int i;
   string acc= get_accents (s);
   s= get_unaccented (s);
   get_extents (s, ex);
@@ -410,12 +526,12 @@ tex_font_rep::accented_get_extents (string s, metric& ex) {
 }
 
 void
-tex_font_rep::accented_get_xpositions (string s, SI* xpos) {
+tex_font_rep::accented_get_xpositions (string s, SI* xpos, bool ligf) {
   int old_status= status;
   status= TEX_ANY;
   string acc= get_accents (s);
   s= get_unaccented (s);
-  get_xpositions (s, xpos);
+  get_xpositions (s, xpos, ligf);
   status= old_status;
 }
 
@@ -424,7 +540,7 @@ tex_font_rep::accented_draw (renderer ren, string s, SI x, SI y) {
   int old_status= status;
   status= TEX_ANY;
 
-  register int i;
+  int i;
   string acc= get_accents (s);
   s= get_unaccented (s);
   draw_fixed (ren, s, x, y);
@@ -485,6 +601,7 @@ tex_font_rep::supports (string s) {
       else return false;
     case TEX_EC:
     case TEX_LA:
+    case TEX_GR:
       return N(s) == 1 || s == "<less>" || s == "<gtr>";
     case TEX_CM:
     case TEX_ADOBE:
@@ -500,12 +617,13 @@ tex_font_rep::supports (string s) {
 
 void
 tex_font_rep::get_extents (string s, metric& ex) {
-  register int i;
+  int i;
   switch (status) {
     case TEX_ANY:
       break;
     case TEX_EC:
     case TEX_LA:
+    case TEX_GR:
       for (i=0; i<N(s); i++)
         if (s[i]=='<') {
           special_get_extents (s, ex);
@@ -535,9 +653,10 @@ tex_font_rep::get_extents (string s, metric& ex) {
   STACK_NEW_ARRAY (ker, int, m);
 
   if (exec) {
-  for (i=0; i<n; i++) s_copy[i]= ((QN) s[i]);
-  tfm->execute (s_copy, n, buf, ker, m);
-  } else {
+    for (i=0; i<n; i++) s_copy[i]= ((QN) s[i]);
+    tfm->execute (s_copy, n, buf, ker, m);
+  }
+  else {
     m = n;
     for (i=0; i<m; ++i) {
       buf[i]= s[i] & 255;
@@ -590,8 +709,9 @@ tex_font_rep::get_extents (string s, metric& ex) {
 }
 
 void
-tex_font_rep::get_xpositions (string s, SI* xpos) {
-  register int i, n= N(s);
+tex_font_rep::get_xpositions (string s, SI* xpos, bool ligf) {
+  int i, n= N(s);
+  xpos[0]= 0;
   if (n == 0) return;
   
   switch (status) {
@@ -599,9 +719,10 @@ tex_font_rep::get_xpositions (string s, SI* xpos) {
       break;
     case TEX_EC:
     case TEX_LA:
+    case TEX_GR:
       for (i=0; i<n; i++)
         if (s[i]=='<') {
-          special_get_xpositions (s, xpos);
+          special_get_xpositions (s, xpos, ligf);
           return;
         }
       break;
@@ -609,12 +730,12 @@ tex_font_rep::get_xpositions (string s, SI* xpos) {
     case TEX_ADOBE:
       for (i=0; i<n; i++) {
         if (s[i]=='<') {
-          special_get_xpositions (s, xpos);
+          special_get_xpositions (s, xpos, ligf);
           return;
         }
         if ((s[i] & 128) != 0) {
           ACCENTS_PREPARE;
-          accented_get_xpositions (s, xpos);
+          accented_get_xpositions (s, xpos, ligf);
           return;
         }
       }
@@ -623,18 +744,24 @@ tex_font_rep::get_xpositions (string s, SI* xpos) {
 
   STACK_NEW_ARRAY (s_copy, int, n);
   for (i=0; i<n; i++) s_copy[i]= ((QN) s[i]);
-  tfm->get_xpositions (s_copy, n, unit, xpos);
+  tfm->get_xpositions (s_copy, n, unit, xpos, ligf);
   STACK_DELETE_ARRAY (s_copy);
 }
 
+void
+tex_font_rep::get_xpositions (string s, SI* xpos) {
+  get_xpositions (s, xpos, true);
+}
+
 void
 tex_font_rep::draw_fixed (renderer ren, string s, SI ox, SI y) {
-  register int i;
+  int i;
   switch (status) {
     case TEX_ANY:
       break;
     case TEX_EC:
     case TEX_LA:
+    case TEX_GR:
       for (i=0; i<N(s); i++)
         if (s[i]=='<') {
           special_draw (ren, s, ox, y);
@@ -665,9 +792,10 @@ tex_font_rep::draw_fixed (renderer ren, string s, SI ox, SI y) {
   STACK_NEW_ARRAY (ker, int, m);
 
   if (exec) {
-  for (i=0; i<n; i++) str[i]= ((QN) s[i]);
-  tfm->execute (str, n, buf, ker, m);
-  } else {
+    for (i=0; i<n; i++) str[i]= ((QN) s[i]);
+    tfm->execute (str, n, buf, ker, m);
+  }
+  else {
     m = n;
     for (i=0; i<m; ++i) {
       buf[i]= s[i] & 255;
@@ -676,7 +804,7 @@ tex_font_rep::draw_fixed (renderer ren, string s, SI ox, SI y) {
   }
 
   for (i=0; i<m; i++) {
-    register int c= buf[i];
+    int c= buf[i];
     glyph gl= pk->get (c);
     if (is_nil (gl)) continue;
     ren->draw (c, pk, x, y);
@@ -688,8 +816,9 @@ tex_font_rep::draw_fixed (renderer ren, string s, SI ox, SI y) {
 }
 
 font
-tex_font_rep::magnify (double zoom) {
-  int ndpi= (int) tm_round (dpi * zoom);
+tex_font_rep::magnify (double zoomx, double zoomy) {
+  if (zoomx != zoomy) return poor_magnify (zoomx, zoomy);
+  int ndpi= (int) tm_round (dpi * zoomx);
   switch (status) {
   case TEX_ANY:
     return tex_font (family, size, ndpi, dsize);
@@ -697,6 +826,8 @@ tex_font_rep::magnify (double zoom) {
     return tex_ec_font (family, size, ndpi, dsize);
   case TEX_LA:
     return tex_la_font (family, size, ndpi, dsize);
+  case TEX_GR:
+    return tex_gr_font (family, size, ndpi, dsize);
   case TEX_CM:
     return tex_cm_font (family, size, ndpi, dsize);
   case TEX_ADOBE:
@@ -713,6 +844,7 @@ tex_font_rep::get_left_correction (string s) {
     break;
   case TEX_EC:
   case TEX_LA:
+  case TEX_GR:
     if (s[0] == '<') return special_get_left_correction (s);
     break;
   case TEX_CM:
@@ -734,6 +866,7 @@ tex_font_rep::get_right_correction (string s) {
     break;
   case TEX_EC:
   case TEX_LA:
+  case TEX_GR:
     if (s[N(s)-1] == '>') return special_get_right_correction (s);
     break;
   case TEX_CM:
@@ -747,16 +880,101 @@ tex_font_rep::get_right_correction (string s) {
   return conv (tfm->i ((QN) s[N(s)-1]));
 }
 
+SI
+tex_font_rep::get_lsub_correction (string s) {
+  SI r= -get_left_correction (s);
+  if (lsub_correct->contains (s)) r += (SI) (lsub_correct[s] * wfn);
+  else if (N(s) > 1 && is_alpha (s[0]) &&
+           lsub_correct->contains (s (0, 1)))
+    r += (SI) (lsub_correct[s (0, 1)] * wfn);
+  return r;
+}
+
+SI
+tex_font_rep::get_lsup_correction (string s) {
+  SI r= 0;
+  if (lsup_correct->contains (s)) r += (SI) (lsup_correct[s] * wfn);
+  else if (N(s) > 1 && is_alpha (s[0]) &&
+           lsup_correct->contains (s (0, 1)))
+    r += (SI) (lsup_correct[s (0, 1)] * wfn);
+  return r;
+}
+
+SI
+tex_font_rep::get_rsub_correction (string s) {
+  SI r= 0;
+  if (rsub_correct->contains (s)) r += (SI) (rsub_correct[s] * wfn);
+  else if (N(s) > 1 && is_alpha (s[N(s)-1]) &&
+           rsub_correct->contains (s (N(s)-1, N(s))))
+    r += (SI) (rsub_correct[s (N(s)-1, N(s))] * wfn);
+  return r;
+}
+
+SI
+tex_font_rep::get_rsup_correction (string s) {
+  SI r= get_right_correction (s);
+  if (rsup_correct->contains (s)) r += (SI) (rsup_correct[s] * wfn);
+  else if (N(s) > 1 && is_alpha (s[N(s)-1]) &&
+           rsup_correct->contains (s (N(s)-1, N(s))))
+    r += (SI) (rsup_correct[s (N(s)-1, N(s))] * wfn);
+  return r;
+}
+
+SI
+tex_font_rep::get_wide_correction (string s, int mode) {
+  if (mode > 0 && above_correct->contains (s))
+    return (SI) (above_correct[s] * wfn);
+  else if (mode < 0 && below_correct->contains (s))
+    return (SI) (below_correct[s] * wfn);
+  else return 0;
+}
+
+void
+tex_font_rep::advance_glyph (string s, int& pos, bool ligf) {
+  if (pos >= N(s)) return;
+  if (!exec || (status != TEX_ANY && s[pos] == '<') || !ligf)
+    tm_char_forwards (s, pos);
+  else {
+    int c= -1;
+    int start= pos;
+    while (pos < N(s)) {
+      int prev= pos;
+      tm_char_forwards (s, pos);
+      string r= s (start, pos);
+
+      int n= N(r);
+      int m= (n+16) << 1;
+      STACK_NEW_ARRAY (str, int, n);
+      STACK_NEW_ARRAY (buf, int, m);
+      STACK_NEW_ARRAY (ker, int, m);
+      for (int i=0; i<n; i++) str[i]= ((QN) r[i]);
+      tfm->execute (str, n, buf, ker, m);
+      bool done= (m > 0 && buf[0] == c);
+      if (!done && m > 0) c= buf[0];
+      STACK_DELETE_ARRAY (str);
+      STACK_DELETE_ARRAY (buf);
+      STACK_DELETE_ARRAY (ker);
+
+      if (done) {
+        pos= prev;
+        break;
+      }
+    }
+  }
+}
+
 glyph
 tex_font_rep::get_glyph (string s) {
-  register int i;
+  int i;
   switch (status) {
   case TEX_ANY:
     break;
   case TEX_EC:
   case TEX_LA:
+  case TEX_GR:
     if (s == "<less>") s= "<";
-    if (s == "<gtr>") s= ">";
+    else if (s == "<gtr>") s= ">";
+    else if (N(s) >= 2 && s[0] == '<') s= special_translate[s];
     break;
   case TEX_CM:
   case TEX_ADOBE:
@@ -767,15 +985,111 @@ tex_font_rep::get_glyph (string s) {
 	return font_rep::get_glyph (s);
     break;
   }
-  if (N(s)!=1) return font_rep::get_glyph (s);
-  int c= ((QN) s[0]);
+  int c;
+  if (N(s) != 1) c= get_ligature_code (s);
+  else c= ((QN) s[0]);
+  if (c == -1) return font_rep::get_glyph (s);
   glyph gl= pk->get (c);
   if (is_nil (gl)) return font_rep::get_glyph (s);
   return gl;
 }
 
+int
+tex_font_rep::index_glyph (string s, font_metric& rm, font_glyphs& rg) {
+  int i;
+  switch (status) {
+  case TEX_ANY:
+    break;
+  case TEX_EC:
+  case TEX_LA:
+  case TEX_GR:
+    if (s == "<less>") s= "<";
+    else if (s == "<gtr>") s= ">";
+    else if (N(s) >= 2 && s[0] == '<') s= special_translate[s];
+    break;
+  case TEX_CM:
+  case TEX_ADOBE:
+    if (s == "<less>") s= "<";
+    if (s == "<gtr>") s= ">";
+    for (i=0; i<N(s); i++)
+      if ((s[i] & 128) != 0 && s[i] != ((char) 254))
+        return font_rep::index_glyph (s, rm, rg);
+    break;
+  }
+  int c;
+  if (N(s) != 1) c= get_ligature_code (s);
+  else c= ((QN) s[0]);
+  if (c == -1) return font_rep::index_glyph (s, rm, rg);
+  glyph gl= pk->get (c);
+  if (is_nil (gl)) {
+    if (c == ((QN) 254)) return -1;
+    return font_rep::index_glyph (s, rm, rg);
+  }
+  rm= tfm_font_metric (tfm, pk, unit);
+  rg= pk;
+  return c;
+}
+
+int
+tex_font_rep::get_ligature_code (string s) {
+  int n= N(s);
+  int m= (n+16) << 1;
+  STACK_NEW_ARRAY (str, int, n);
+  STACK_NEW_ARRAY (buf, int, m);
+  STACK_NEW_ARRAY (ker, int, m);
+  for (int i=0; i<n; i++) str[i]= ((QN) s[i]);
+  tfm->execute (str, n, buf, ker, m);
+  STACK_DELETE_ARRAY (str);
+  STACK_DELETE_ARRAY (buf);
+  STACK_DELETE_ARRAY (ker);
+  if (m == 1) return buf[0];
+  else return -1;
+}
+
+/******************************************************************************
+* TeX font metrics as usual metrics
+******************************************************************************/
+
+extern metric error_metric;
+
+struct tfm_font_metric_rep: public font_metric_rep {
+  tex_font_metric tfm;
+  font_glyphs pk;
+  double unit;
+  hashmap<int,pointer> ms;
+  tfm_font_metric_rep (string name, tex_font_metric tfm2,
+                       font_glyphs pk2, double unit2):
+    font_metric_rep (name), tfm (tfm2), pk (pk2),
+    unit (unit2), ms (error_metric) {}
+  bool exists (int c) {
+    return c >= ((int) tfm->bc) && ((int) tfm->ec) >= c; }
+  metric& get (int c) {
+    if (!exists (c)) return error_metric;
+    if (!ms->contains (c)) {
+      metric_struct* r= tm_new<metric_struct> ();
+      ms(c)= (pointer) r;
+      r->x1=  0;
+      r->x2=  conv (tfm->w(c));
+      r->y1= -conv (tfm->d(c));
+      r->y2=  conv (tfm->h(c));
+      glyph gl= pk->get (c);
+      r->x3= -((int) gl->xoff) * PIXEL;
+      r->x4=  ((int) (gl->width- gl->xoff)) * PIXEL;
+      r->y3=  ((int) (gl->yoff- gl->height)) * PIXEL;
+      r->y4=  ((int) gl->yoff) * PIXEL;
+    }
+    return *((metric*) ((void*) ms[c])); }
+};
+
 #undef conv
 
+font_metric
+tfm_font_metric (tex_font_metric tfm, font_glyphs pk, double unit) {
+  string name= tfm->res_name * ":" * pk->res_name * ":" * as_string (unit);
+  return make (font_metric, name,
+               tm_new<tfm_font_metric_rep> (name, tfm, pk, unit));
+}
+
 /******************************************************************************
 * Interface
 ******************************************************************************/
@@ -808,6 +1122,13 @@ tex_la_font (string family, int size, int dpi, int dsize) {
     tm_new<tex_font_rep> (name, TEX_LA, family, size, dpi, dsize));
 }
 
+font
+tex_gr_font (string family, int size, int dpi, int dsize) {
+  string name= "gr:" * family * as_string (size) * "@" * as_string(dpi);
+  return make (font, name,
+    tm_new<tex_font_rep> (name, TEX_GR, family, size, dpi, dsize));
+}
+
 font
 tex_adobe_font (string family, int size, int dpi, int dsize) {
   string name= "adobe:" * family * as_string (size) * "@" * as_string(dpi);
