@@ -21,6 +21,7 @@
 #include <QObject>
 #include <QWidget>
 #include <QPaintDevice>
+#include <QPainterPath>
 #include <QPixmap>
 
 /******************************************************************************
@@ -78,12 +79,23 @@ static hashmap<basic_character,qt_image> character_image;
 // image cache
 static hashmap<string,qt_pixmap> images;
 
+/*
+** hash contents must be removed because 
+** the underlying objects are destroyed during 
+** Qt exit function
+*/
+void del_obj_qt_renderer(void)  {
+  character_image= hashmap<basic_character,qt_image> ();  
+  images= hashmap<string,qt_pixmap>() ;
+}
+
 /******************************************************************************
 * qt_renderer
 ******************************************************************************/
 
 qt_renderer_rep::qt_renderer_rep (QPainter *_painter, int w2, int h2):
-  basic_renderer_rep (true, w2, h2), painter(_painter) {}
+  basic_renderer_rep (true, w2, h2), painter(_painter) {
+    reset_zoom_factor(); }
 
 qt_renderer_rep::~qt_renderer_rep () {}
 
@@ -96,9 +108,9 @@ void
 qt_renderer_rep::begin (void* handle) {
   QPaintDevice *device = static_cast<QPaintDevice*>(handle);
   if (!painter->begin (device) && DEBUG_QT)
-    cout << "qt_renderer_rep::begin(): uninitialized QPixmap of size "
-         << ((QPixmap*)handle)->width() << " x " << ((QPixmap*)handle)->height()
-         << LF;
+    debug_qt << "qt_renderer_rep::begin(): uninitialized QPixmap of size "
+             << ((QPixmap*)handle)->width() << " x "
+             << ((QPixmap*)handle)->height() << LF;
     
   w = painter->device()->width();
   h = painter->device()->height();
@@ -115,6 +127,12 @@ qt_renderer_rep::get_extents (int& w2, int& h2) {
   }
 }
 
+void
+qt_renderer_rep::set_zoom_factor (double zoom) {
+  renderer_rep::set_zoom_factor (retina_factor * zoom);
+  retina_pixel= pixel * retina_factor;
+}
+
 /******************************************************************************
 * Transformations
 ******************************************************************************/
@@ -170,49 +188,34 @@ qt_renderer_rep::set_clipping (SI x1, SI y1, SI x2, SI y2, bool restore)
 /******************************************************************************
 * Drawing 
 ******************************************************************************/
-
 bool is_percentage (tree t, string s= "%");
 double as_percentage (tree t);
 
 static QImage*
 get_pattern_image (brush br, SI pixel) {
-  tree pattern= br->get_pattern ();
-  url u= as_string (pattern[0]);
-  int imw_pt, imh_pt;
-  image_size (u, imw_pt, imh_pt);
-  double pt= ((double) 600*PIXEL) / 72.0;
-  SI imw= (SI) (((double) imw_pt) * pt);
-  SI imh= (SI) (((double) imh_pt) * pt);
-
-  SI w= imw, h= imh;
-  if (is_int (pattern[1])) w= as_int (pattern[1]);
-  else if (is_percentage (pattern[1]))
-    w= (SI) (as_percentage (pattern[1]) * ((double) w));
-  else if (is_percentage (pattern[1], "@"))
-    w= (SI) (as_percentage (pattern[1]) * ((double) h));
-  if (is_int (pattern[2])) h= as_int (pattern[2]);
-  else if (is_percentage (pattern[2]))
-    h= (SI) (as_percentage (pattern[2]) * ((double) h));
-  else if (is_percentage (pattern[2], "@"))
-    h= (SI) (as_percentage (pattern[2]) * ((double) w));
-  w= ((w + pixel - 1) / pixel);
-  h= ((h + pixel - 1) / pixel);
-  QImage* pm= get_image (u, w, h);
+  url u;
+  SI w, h;
+  tree eff;
+  get_pattern_data (u, w, h, eff, br, pixel);
+  QImage* pm= get_image (u, w, h, eff, pixel);
   return pm;
 }
 
 void
 qt_renderer_rep::set_pencil (pencil np) {
+  painter->setOpacity (qreal (1.0));
   basic_renderer_rep::set_pencil (np);
   QPen p (painter->pen ());
   QBrush b (painter->brush ());
   QColor qc= to_qcolor (pen->get_color ());
-  SI pw= 0;
-  if (pen->get_width () > pixel)
-    pw= (pen->get_width () + thicken) / (1.0*pixel);
   p.setColor (qc);
   b.setColor (qc);
-  p.setWidth (pw);
+  //SI pw= 0;
+  //if (pen->get_width () > pixel)
+  //pw= (pen->get_width () + thicken) / (1.0*pixel);
+  //p.setWidth (pw);
+  qreal pw= (qreal) (((double) pen->get_width ()) / ((double) pixel));
+  p.setWidthF (pw);
   if (np->get_type () == pencil_brush) {
     brush br= np->get_brush ();
     QImage* pm= get_pattern_image (br, pixel);
@@ -220,9 +223,15 @@ qt_renderer_rep::set_pencil (pencil np) {
     painter->setOpacity (qreal (pattern_alpha) / qreal (255));
     if (pm != NULL) {
       b= QBrush (*pm);
+      double pox, poy;
+      decode (0, 0, pox, poy);
+      QTransform tr;
+      tr.translate (pox, poy);
+      b.setTransform (tr);
       p= QPen (b, pw);
     }
   }
+  p.setStyle (Qt::SolidLine);
   p.setCapStyle (pen->get_cap () == cap_round? Qt::RoundCap: Qt::SquareCap);
   p.setJoinStyle (Qt::RoundJoin);
   painter->setPen (p);
@@ -249,35 +258,55 @@ qt_renderer_rep::set_brush (brush br) {
     QImage* pm= get_pattern_image (br, pixel);
     int pattern_alpha= br->get_alpha ();
     painter->setOpacity (qreal (pattern_alpha) / qreal (255));
-    if (pm != NULL) painter->setBrush (QBrush (*pm));
+    if (pm != NULL) {
+      QBrush b (*pm);
+      double pox, poy;
+      decode (0, 0, pox, poy);
+      QTransform tr;
+      tr.translate (pox, poy);
+      //tr.rotate (45.0);
+      b.setTransform (tr);
+      painter->setBrush (b);
+    }
   }
 }
 
 void
 qt_renderer_rep::line (SI x1, SI y1, SI x2, SI y2) {
-  decode (x1, y1);
-  decode (x2, y2);
+  double rx1, ry1, rx2, ry2;
+  decode (x1, y1, rx1, ry1);
+  decode (x2, y2, rx2, ry2);
   // y1--; y2--; // top-left origin to bottom-left origin conversion
   painter->setRenderHints (QPainter::Antialiasing);
-  painter->drawLine (x1, y1, x2, y2);
+  painter->drawLine (QPointF (rx1, ry1), QPointF (rx2, ry2));
 }
 
 void
 qt_renderer_rep::lines (array<SI> x, array<SI> y) {
   int i, n= N(x);
-  if ((N(y) != n) || (n<1)) return;
-  STACK_NEW_ARRAY (pnt, QPoint, n);
-  for (i=0; i<n; i++) {
-    SI xx= x[i], yy= y[i];
-    decode (xx, yy);
-    pnt[i].rx()= xx;
-    pnt[i].ry()= yy;
-    if (i>0) {
-      painter->setRenderHints (QPainter::Antialiasing);
-      painter->drawLine (pnt[i-1], pnt[i]); // FIX: hack
-    }
+  if ((N(y) != n) || (n<2)) return;
+  STACK_NEW_ARRAY (pnt, QPointF, n);
+  for (i=0; i<n; i++)
+#ifdef __arm__
+// for compiling in Raspbian
+  {
+    double px = double (pnt[i].rx());//explicit conversion needed for ARM where qreal==float!=double
+    double py = double (pnt[i].ry());    
+    decode (x[i], y[i], px, py);
+    pnt[i].rx()=qreal(px); pnt[i].ry()=qreal(py);
   }
-  // XDrawLines (dpy, win, gc, pnt, n, CoordModeOrigin);
+#else
+    decode (x[i], y[i], pnt[i].rx(), pnt[i].ry());
+#endif
+
+  QPen p= painter->pen();
+  p.setCapStyle (pen->get_cap () == cap_round? Qt::RoundCap: Qt::SquareCap);
+  if (x[N(x)-1] == x[0] && y[N(y)-1] == y[0]) p.setCapStyle (Qt::RoundCap);
+  p.setJoinStyle (Qt::RoundJoin);
+  painter->setPen (p);
+
+  painter->setRenderHints (QPainter::Antialiasing);
+  painter->drawPolyline (pnt, n);
   STACK_DELETE_ARRAY (pnt);
 }
 
@@ -323,21 +352,25 @@ qt_renderer_rep::fill (SI x1, SI y1, SI x2, SI y2) {
 void
 qt_renderer_rep::arc (SI x1, SI y1, SI x2, SI y2, int alpha, int delta) {
   if ((x1>=x2) || (y1>=y2)) return;
-  decode (x1, y1);
-  decode (x2, y2);
+  double rx1, ry1, rx2, ry2;
+  decode (x1, y1, rx1, ry1);
+  decode (x2, y2, rx2, ry2);
   painter->setRenderHints (QPainter::Antialiasing);
-  painter->drawArc (x1, y2, x2-x1, y1-y2, alpha / 4, delta / 4);
+  painter->drawArc (QRectF (rx1, ry2, rx2-rx1, ry1-ry2), alpha / 4, delta / 4);
 }
 
 void
 qt_renderer_rep::fill_arc (SI x1, SI y1, SI x2, SI y2, int alpha, int delta) {
   if ((x1>=x2) || (y1>=y2)) return;
-  decode (x1, y1);
-  decode (x2, y2);
-  QBrush br(to_qcolor (pen->get_color ()));
+  double rx1, ry1, rx2, ry2;
+  decode (x1, y1, rx1, ry1);
+  decode (x2, y2, rx2, ry2);
+  QBrush br= painter->brush ();
+  if (is_nil (fg_brush) || fg_brush->get_type () != brush_pattern)
+    br= QBrush (to_qcolor (pen->get_color ()));
   QPainterPath pp;
-  pp.arcMoveTo (x1, y2, x2-x1, y1-y2, alpha / 64);
-  pp.arcTo (x1, y2, x2-x1, y1-y2, alpha / 64, delta / 64);
+  pp.arcMoveTo (QRectF (rx1, ry2, rx2-rx1, ry1-ry2), alpha / 64);
+  pp.arcTo (QRectF (rx1, ry2, rx2-rx1, ry1-ry2), alpha / 64, delta / 64);
   pp.closeSubpath ();
   pp.setFillRule (Qt::WindingFill);
   painter->setRenderHints (QPainter::Antialiasing);
@@ -348,11 +381,11 @@ void
 qt_renderer_rep::polygon (array<SI> x, array<SI> y, bool convex) {
   int i, n= N(x);
   if ((N(y) != n) || (n<1)) return;
-  QPolygonF poly(n);
+  QPolygonF poly (n);
   for (i=0; i<n; i++) {
-    SI xx= x[i], yy= y[i];
-    decode (xx, yy);
-    poly[i] = QPointF (xx, yy);
+    double qx, qy;
+    decode (x[i], y[i], qx, qy);
+    poly[i] = QPointF (qx, qy);
   }
   QBrush br= painter->brush ();
   if (is_nil (fg_brush) || fg_brush->get_type () != brush_pattern)
@@ -381,6 +414,9 @@ qt_renderer_rep::draw_triangle (SI x1, SI y1, SI x2, SI y2, SI x3, SI y3) {
     SI xx= x[i], yy= y[i];
     decode (xx, yy);
     poly[i] = QPointF (xx, yy);
+    //double qx, qy;
+    //decode (x[i], y[i], qx, qy);
+    //poly[i] = QPointF (qx, qy);
   }
   QBrush br= painter->brush ();
   if (is_nil (fg_brush) || fg_brush->get_type () != brush_pattern)
@@ -455,6 +491,7 @@ qt_renderer_rep::draw_bis (int c, font_glyphs fng, SI x, SI y) {
         color patcol= im->pixel (i, j);
         int r, g, b, a;
         get_rgb (patcol, r, g, b, a);
+        if (get_reverse_colors ()) reverse (r, g, b);
         int col = gl->get_x (i, j);
         im->setPixel (i, j, qRgba (r, g, b, (a*col)/nr_cols));
       }
@@ -478,6 +515,7 @@ qt_renderer_rep::draw (int c, font_glyphs fng, SI x, SI y) {
   if (is_nil(mi)) {
     int r, g, b, a;
     get_rgb (fgc, r, g, b, a);
+    if (get_reverse_colors ()) reverse (r, g, b);
     SI xo, yo;
     glyph pre_gl= fng->get (c); if (is_nil (pre_gl)) return;
     glyph gl= shrink (pre_gl, std_shrinkf, std_shrinkf, xo, yo);
