@@ -21,14 +21,14 @@
 * There are three main types of urls:
 *   - rootless urls, like a/b/c. These urls are mainly used in computations.
 *     For example, they can be appended to another url.
-*   - Standard rooted urls, like file:///usr or http://www.texmacs.org.
+*   - Standard rooted urls, like file:///usr or https://www.texmacs.org.
 *     These are the same as those used on the web.
 *   - System urls, characterized by a "default" root.
 *     These urls are similar to standard rooted urls, but they behave
 *     in a slightly different way with respect to concatenation.
-*     For instance http://www.texmacs.org/Web * file:///tmp would yield
-*     file:///tmp, where as http://www.texmacs.org/Web * /tmp yields
-*     http://www.texmacs.org/tmp
+*     For instance https://www.texmacs.org/Web * file:///tmp would yield
+*     file:///tmp, where as https://www.texmacs.org/Web * /tmp yields
+*     https://www.texmacs.org/tmp
 *******************************************************************************
 * There are several formats for parsing (and printing) urls:
 *   - System format: the usual format on your operating system.
@@ -67,7 +67,7 @@
 
 #include <ctype.h>
 
-#if defined(OS_WIN32)  || defined(__MINGW32__)
+#ifdef OS_MINGW
 #define WINPATHS
 #endif
 
@@ -94,6 +94,10 @@ url_get_atom (string s, int type) {
     }
   }
   if (occurs ("*", s)) return url_wildcard (s);
+#ifdef WINPATHS
+  if (N(s)==2 && ends (s, ":"))
+    s->resize(1); // remove the ':' after unit letter
+#endif
   return as_url (tree (s));
 }
 
@@ -266,8 +270,10 @@ url_general (string name, int type= URL_SYSTEM) {
   if (heuristic_is_path (name, type)) return url_path (name, type);
   if (heuristic_is_default (name, type)) return url_default (name, type);
   if (heuristic_is_mingw_default (name, type)) return url_mingw_default (name, type);
-  if (heuristic_is_http (name)) return url_http (name);
-  if (heuristic_is_ftp (name)) return url_ftp (name);
+  if (type != URL_CLEAN_UNIX) {
+    if (heuristic_is_http (name)) return url_http (name);
+    if (heuristic_is_ftp (name)) return url_ftp (name);
+  }
   return url_get_name (name, type);
 }
 
@@ -429,6 +435,22 @@ is_rooted_tmfs (url u) {
     (is_or (u) && is_rooted_tmfs (u[1]) && is_rooted_tmfs (u[2]));
 }
 
+bool
+is_tmfs_protocol (url u, string protocol) {
+  return
+    u->t == protocol ||
+    (is_concat (u) && is_tmfs_protocol (u[1], protocol));
+}
+
+bool
+is_rooted_tmfs (url u, string protocol) {
+  return
+    (is_concat (u) && is_root_tmfs (u[1]) &&
+                      is_tmfs_protocol (u[2], protocol)) ||
+    (is_or (u) && is_rooted_tmfs (u[1], protocol) &&
+                  is_rooted_tmfs (u[2], protocol));
+}
+
 bool
 is_rooted_blank (url u) {
   return
@@ -493,23 +515,27 @@ as_string (url u, int type) {
     if ((!is_concat (u[2])) && (!is_atomic (u[2])) && (!is_wildcard (u[2], 1)))
       s2= "{" * s2 * "}";
 #ifdef WINPATHS
-    if (is_semi_root (u)) {
-      if (ends (s2, ":")) {
-        if (type == URL_SYSTEM) return s2 * "\\";
-        else return s1 * sep * s2 (0, N(s2) - 1);
+    if (((is_root (u[1], "default") && type == URL_SYSTEM) ||
+         is_root (u[1], "file"))) { // have to return the windows format
+      string root, remain;
+      if (is_concat (u[2])) {           
+        root = as_string (u[2][1], type);
+        // root might be unit letter or hostname. It depends on the length
+        remain = as_string (u[2][2], type);
       }
       else {
-        if (type == URL_SYSTEM) return "\\\\" * s2;
-        else return s1 * sep * s2;
+        root = s2;
+        remain = "";
+      }
+      if (is_root (u[1], "default")) {
+        if (N(root) == 1) return root * ":\\" * remain; // drive letter
+        else return "\\\\" * root * "\\" * remain;
+      }
+      else {
+        if (N(root) == 1) return s1 * "/" * root * ":/" * remain; // local file
+        else return s1 * root * "/" * remain; // remote
       }
     }
-    if (is_root (u[1]) && type != URL_SYSTEM && N(s2) >= 2 && is_alpha (s2[0]) && s2[1] == ':')
-      return s1 * sep * s2 (0, 1) * s2 (2, N(s2));
-    if (is_root (u[1]) && stype == URL_SYSTEM) {
-      if (N(s2) >= 2 && is_alpha (s2[0]) && s2[1] == ':') return s2; 
-      else return "\\\\" * s2; 
-    }
-
 #endif
     return s1 * sep * s2;
   }
@@ -529,7 +555,7 @@ as_string (url u, int type) {
   if (is_root (u, "default")) {
     int stype= type;
     if (is_root (u[1]) && (!is_root (u[1], "default"))) stype= URL_STANDARD;
-	if (stype == URL_SYSTEM) return ""; else return "/";
+        if (stype == URL_SYSTEM) return ""; else return "/";
   }
 #else
   if (is_root (u, "default")) return "/";
@@ -579,7 +605,7 @@ suffix (url u) {
   if ((i>0) && (i<n-1)) {
     string r= s (i+1, n);
     while ((N(r)>0) && (r[N(r)-1]=='~' || r[N(r)-1]=='#')) r= r(0, N(r)-1);
-    return r;
+    return locase_all(r);
   }
   return "";
 }
@@ -604,6 +630,7 @@ glue (url u, string s) {
   if (is_atomic (u)) return as_url (tree (u->t->label * s));
   if (is_concat (u)) return u[1] * glue (u[2], s);
   if (is_or (u)) return glue (u[1], s) | glue (u[2], s);
+  if (is_none (u)) return u;
   failed_error << "u= " << u << "\n";
   failed_error << "s= " << s << "\n";
   FAILED ("can't glue string to url");
@@ -616,6 +643,7 @@ unglue (url u, int nr) {
     return as_url (tree (u->t->label (0, max (N(u->t->label) - nr, 0))));
   if (is_concat (u)) return u[1] * unglue (u[2], nr);
   if (is_or (u)) return unglue (u[1], nr) | unglue (u[2], nr);
+  if (is_none (u)) return u;
   failed_error << "u = " << u << "\n";
   failed_error << "nr= " << nr << "\n";
   FAILED ("can't unglue from url");
@@ -641,8 +669,8 @@ delta_sub (url base, url u) {
     return u;
   if (is_concat (base) && is_concat (u) && (base[1] == u[1])) {
     if (is_special_root (base[1]) &&
-	is_concat (base[2]) && is_concat (u[2]) &&
-	base[2][1] != u[2][1])
+        is_concat (base[2]) && is_concat (u[2]) &&
+        base[2][1] != u[2][1])
       return url_none ();
     return delta_sub (base[2], u[2]);
   }
@@ -683,7 +711,12 @@ expand (url u) {
 
 bool
 descends (url u, url base) {
-  if (is_or (base)) return descends (u, base[1]) || descends (u, base[2]);
+  if (is_or (base))
+    return descends (u, base[1]) || descends (u, base[2]);
+  if (is_or (u))
+    return descends (u[1], base) && descends (u[2], base);
+  if (u == base)
+    return true;
   if (is_concat (u) && is_atomic (base))
     return u[1] == base;
   if (is_concat (u) && is_concat (base))
@@ -812,8 +845,10 @@ complete (url base, url u, string filter, bool flag) {
      if (is_none (base)) return base;
      if (is_none (u)) return u;
      if ((!is_root (base)) && (!is_rooted_name (base))) {
-        failed_error << "base= " << base << LF;
-        FAILED ("invalid base url");
+       failed_error << "base  = " << base << LF;
+       failed_error << "u     = " << u << LF;
+       failed_error << "filter= " << filter << LF;
+       FAILED ("invalid base url");
      }
   }
   if (is_name (u) || (is_concat (u) && is_root (u[1]) && is_name (u[2]))) {
@@ -826,8 +861,9 @@ complete (url base, url u, string filter, bool flag) {
       if (is_of_type (comp, filter)) return u;
       return url_none ();
     }
-    failed_error << "base= " << base << LF;
-    failed_error << "u= " << u << LF;
+    failed_error << "base  = " << base << LF;
+    failed_error << "u     = " << u << LF;
+    failed_error << "filter= " << filter << LF;
     ASSERT (is_rooted (comp), "unrooted url");
     FAILED ("bad protocol in url");
   }
@@ -838,7 +874,9 @@ complete (url base, url u, string filter, bool flag) {
   if (is_concat (u) && is_wildcard (u[1], 0) && is_wildcard (u[2], 1)) {
     // FIXME: ret= ret | ... is unefficient (quadratic) in main loop
     if (!(is_rooted (base, "default") || is_rooted (base, "file"))) {
-      failed_error << "base= " << base << LF;
+      failed_error << "base  = " << base << LF;
+      failed_error << "u     = " << u << LF;
+      failed_error << "filter= " << filter << LF;
       FAILED ("wildcards only implemented for files");
     }
     url ret= url_none ();
@@ -848,11 +886,13 @@ complete (url base, url u, string filter, bool flag) {
     for (i=0; i<n; i++) {
       if ((!is_none (ret)) && flag) return ret;
       if ((dir[i] == ".") || (dir[i] == "..")) continue;
-      if (starts (dir[i], "http://") || starts (dir[i], "ftp://"))
+      if (starts (dir[i], "http://") ||
+          starts (dir[i], "https://") ||
+          starts (dir[i], "ftp://"))
         if (is_directory (base * dir[i])) continue;
       ret= ret | (dir[i] * complete (base * dir[i], u, filter, flag));
       if (match_wildcard (dir[i], u[2][1]->t->label))
-	ret= ret | complete (base, dir[i], filter, flag);
+        ret= ret | complete (base, dir[i], filter, flag);
     }
     return ret;
   }
@@ -869,7 +909,9 @@ complete (url base, url u, string filter, bool flag) {
   if (is_wildcard (u)) {
     // FIXME: ret= ret | ... is unefficient (quadratic) in main loop
     if (!(is_rooted (base, "default") || is_rooted (base, "file"))) {
-      failed_error << "base= " << base << LF;
+      failed_error << "base  = " << base << LF;
+      failed_error << "u     = " << u << LF;
+      failed_error << "filter= " << filter << LF;
       FAILED ("wildcards only implemented for files");
     }
     url ret= url_none ();
@@ -880,12 +922,14 @@ complete (url base, url u, string filter, bool flag) {
     for (i=0; i<n; i++) {
       if ((!is_none (ret)) && flag) return ret;
       if ((dir[i] == ".") || (dir[i] == "..")) continue;
-      if (starts (dir[i], "http://") || starts (dir[i], "ftp://"))
+      if (starts (dir[i], "http://") ||
+          starts (dir[i], "https://") ||
+          starts (dir[i], "ftp://"))
         if (is_directory (base * dir[i])) continue;
       if (is_wildcard (u, 0))
-	ret= ret | (dir[i] * complete (base * dir[i], u, filter, flag));
+        ret= ret | (dir[i] * complete (base * dir[i], u, filter, flag));
       else if (match_wildcard (dir[i], u[1]->t->label))
-	ret= ret | complete (base, dir[i], filter, flag);
+        ret= ret | complete (base, dir[i], filter, flag);
     }
     return ret;
   }
@@ -931,11 +975,16 @@ resolve_in_path (url u) {
     if (ends (which, name))
       return which;
     else if ((which != "") &&
-	     (!starts (which, "which: ")) &&
-	     (!starts (which, "no ")))
+             (!occurs ("bin/which: ", which)) &&
+             (!starts (which, "which: ")) &&
+             (!starts (which, "no ")))
       cout << "TeXmacs] " << which << "\n";
   }
+#ifdef OS_MINGW
+  return resolve ((url_path ("$TEXMACS_PATH/bin") | url_path ("$PATH")) * u, "x");
+#else
   return resolve (url_path ("$PATH") * u, "x");
+#endif
 }
 
 bool
@@ -945,10 +994,10 @@ exists (url u) {
 
 bool
 exists_in_path (url u) {
-#if defined (OS_WIN32) || defined (__MINGW__) || defined (__MINGW32__)
+#ifdef OS_MINGW
   return !is_none (resolve_in_path (url (as_string (u) * ".bat"))) ||\
-  	 !is_none (resolve_in_path (url (as_string (u) * ".exe"))) ||\
-	 !is_none (resolve_in_path (url (as_string (u) * ".com")));
+         !is_none (resolve_in_path (url (as_string (u) * ".exe"))) ||\
+         !is_none (resolve_in_path (url (as_string (u) * ".com")));
 #else
   return !is_none (resolve_in_path (u));
 #endif
@@ -973,32 +1022,48 @@ descendance (url u) {
   return factor (descendance_sub (u));
 }
 
+url
+subdirectories (url u) {
+  if (is_or (u))
+    return subdirectories (u[1]) | subdirectories (u[2]);
+  else if (is_directory (u)) {
+    url ret= u;
+    bool error_flag;
+    array<string> dir= read_directory (u, error_flag);
+    for (int i=0; i<N(dir); i++)
+      if (!starts (dir[i], ".") && is_directory (u * dir[i]))
+        ret= ret | subdirectories (u * dir[i]);
+    return ret;
+  }
+  else return url_none ();
+}
+
 /******************************************************************************
 * Concretization of resolved urls
 ******************************************************************************/
 
+url
+concretize_url (url u) {
+  // This routine transforms a resolved url into a system url.
+  // In the case of distant files from the web, a local copy is created.
+  if (is_rooted (u, "default") ||
+      is_rooted (u, "file") ||
+      is_rooted (u, "blank"))
+        return reroot (u, "default");
+  if (is_rooted_web (u)) return concretize_url (get_from_web (u));
+  if (is_rooted_tmfs (u)) return concretize_url (get_from_server (u));
+  if (is_ramdisc (u)) return concretize_url (get_from_ramdisc (u));
+  if (is_here (u)) return url_pwd ();
+  if (is_parent (u)) return url_pwd () * url_parent ();
+  return url_none ();
+}
+
 string
 concretize (url u) {
   // This routine transforms a resolved url into a system file name.
   // In the case of distant files from the web, a local copy is created.
-#ifdef WINPATHS
-  // FIXME: this fix seems strange;
-  // to start with, the if condition is not respected
-  string s = as_string (u);
-  if (starts (s, "file:///")) s= s (8, N(s));
-  if (heuristic_is_default (s, 0)) return s;
-  if (is_rooted (u, "blank")) return as_string (reroot (u, "default"));
-#else
-  if (is_rooted (u, "default") ||
-      is_rooted (u, "file") ||
-      is_rooted (u, "blank"))
-    return as_string (reroot (u, "default"));
-#endif
-  if (is_rooted_web (u)) return concretize (get_from_web (u));
-  if (is_rooted_tmfs (u)) return concretize (get_from_server (u));
-  if (is_ramdisc (u)) return concretize (get_from_ramdisc (u));
-  if (is_here (u)) return as_string (url_pwd ());
-  if (is_parent (u)) return as_string (url_pwd () * url_parent ());
+  url c= concretize_url (u);
+  if (!is_none (c)) return as_string (c);
   if (is_wildcard (u, 1)) return u->t[1]->label;
   std_warning << "Couldn't concretize " << u->t << LF;
   // failed_error << "u= " << u << LF;
