@@ -24,16 +24,22 @@
 class dialogue_command_rep: public command_rep {
   server_rep* sv;
   object      fun;
+  scheme_tree p;
   int         nr_args;
 
 public:
   dialogue_command_rep (server_rep* sv2, object fun2, int nr_args2):
     sv (sv2), fun (fun2), nr_args (nr_args2) {}
+  dialogue_command_rep (server_rep* sv2, object fun2, scheme_tree p2):
+    sv (sv2), fun (fun2), p (p2), nr_args (N(p2)) {}
   void apply ();
   tm_ostream& print (tm_ostream& out) {
     return out << "Dialogue"; }
 };
 
+static string
+get_type (scheme_tree p, int i);
+
 void
 dialogue_command_rep::apply () {
   int i;
@@ -47,8 +53,11 @@ dialogue_command_rep::apply () {
       return;
     }
     object arg= string_to_object (s_arg);
-    learn= cons (cons (object (as_string (i)), arg), learn);
     cmd= cons (arg, cmd);
+    if (!is_empty (p) && get_type (p, i) == "password")
+      learn= cons (cons (object (as_string (i)), object ("")), learn);
+    else
+      learn= cons (cons (object (as_string (i)), arg), learn);
     //call ("learn-interactive-arg", fun, object (i), arg);
   }
   call ("learn-interactive", fun, learn);
@@ -58,8 +67,13 @@ dialogue_command_rep::apply () {
 }
 
 command
-dialogue_command (server_rep* sv, object fun, int nr_args) {
-  return tm_new<dialogue_command_rep> (sv, fun, nr_args);
+dialogue_command (server_rep* sv, object fun, scheme_tree p) {
+  return tm_new<dialogue_command_rep> (sv, fun, p);
+}
+
+command
+dialogue_command (server_rep* sv, object fun, int n) {
+  return tm_new<dialogue_command_rep> (sv, fun, n);
 }
 
 void
@@ -71,7 +85,7 @@ tm_frame_rep::dialogue_start (string name, widget wid) {
     dialogue_wid= wid;
     dialogue_win= plain_window_widget (dialogue_wid, name);
 
-    widget win= get_window () -> win;
+    widget win= concrete_window () -> win;
     SI ox, oy, dx, dy, ex= 0, ey= 0;
     get_position (win, ox, oy);
     get_size (win, dx, dy);
@@ -102,31 +116,20 @@ tm_frame_rep::dialogue_end () {
   }
 }
 
+/*
 static int
 gcd (int i, int j) {
   if (i<j)  return gcd (j, i);
   if (j==0) return i;
   return gcd (j, i%j);
 }
+*/
 
 void
-tm_frame_rep::choose_file (object fun, string title, string type) {
-  string magn;
-  if (type == "image") {
-    editor ed   = get_editor ();
-    int dpi     = as_int (ed->get_env_string (DPI));
-    int sfactor = get_window () -> get_shrinking_factor ();
-    int num     = 75*sfactor;
-    int den     = dpi;
-    int g       = gcd (num, den);
-    num /= g; den /= g;
-    if (num != 1) magn << "*" << as_string (num);
-    if (den != 1) magn << "/" << as_string (den);
-  }
-
-  url      name= get_name_buffer ();
+tm_frame_rep::choose_file (object fun, string title, string type,
+			   string prompt, url name) {
   command  cb  = dialogue_command (get_server(), fun, 1);
-  widget   wid = file_chooser_widget (cb, type, magn);
+  widget   wid = file_chooser_widget (cb, type, prompt);
   if (!is_scratch (name)) {
     set_directory (wid, as_string (head (name)));
     if ((type != "image") && (type != "")) {
@@ -134,11 +137,11 @@ tm_frame_rep::choose_file (object fun, string title, string type) {
       string old_suf= suffix (u);
       string new_suf= format_to_suffix (type);
       if ((suffix_to_format (suffix (u)) != type) &&
-	  (old_suf != "") && (new_suf != ""))
-	{
-	  u= unglue (u, N(old_suf) + 1);
-	  u= glue (u, "." * new_suf);
-	}
+          (old_suf != "") && (new_suf != ""))
+        {
+          u= unglue (u, N(old_suf) + 1);
+          u= glue (u, "." * new_suf);
+        }
       set_file (wid, as_string (u));
     }
   }
@@ -155,11 +158,13 @@ tm_frame_rep::choose_file (object fun, string title, string type) {
 static string
 get_prompt (scheme_tree p, int i) {
   if (is_atomic (p[i]) && is_quoted (p[i]->label))
-    return scm_unquote (p[i]->label);
-  else if (is_tuple (p[i]) && N(p[i])>0 &&
-	   is_atomic (p[i][0]) && is_quoted (p[i][0]->label))
-    return scm_unquote (p[i][0]->label);
-  return "Input:";
+    return translate (scm_unquote (p[i]->label));
+  else if (is_tuple (p[i]) && N(p[i])>0) {
+    if (is_atomic (p[i][0]) && is_quoted (p[i][0]->label))
+      return translate (scm_unquote (p[i][0]->label));
+    return translate (scheme_tree_to_tree (p[i][0]));
+  }
+  return translate ("Input:");
 }
 
 static string
@@ -177,7 +182,7 @@ get_proposals (scheme_tree p, int i) {
     int j, n= N(p[i]);
     for (j=2; j<n; j++)
       if (is_atomic (p[i][j]) && is_quoted (p[i][j]->label))
-	a << scm_unquote (p[i][j]->label);
+        a << scm_unquote (p[i][j]->label);
   }
   return a;
 }
@@ -207,12 +212,15 @@ interactive_command_rep::apply () {
     array<object> params (N(p));
     for (i=N(p)-1; i>=0; i--) {
       params[i]= string_to_object (s[i]);
-      learn= cons (cons (object (as_string (i)), params[i]), learn);
+      if (get_type (p, i) == "password")
+        learn= cons (cons (object (as_string (i)), object ("")), learn);
+      else
+        learn= cons (cons (object (as_string (i)), params[i]), learn);
     }
     call ("learn-interactive", fun, learn);
     string ret= object_to_string (call (fun, params));
     if (ret != "" && ret != "<unspecified>" && ret != "#<unspecified>")
-      sv->set_message (ret, "interactive command");
+      sv->set_message (verbatim (ret), "interactive command");
   }
   else {
     s[i]= string ("");
@@ -227,12 +235,20 @@ interactive_command_rep::apply () {
 void
 tm_frame_rep::interactive (object fun, scheme_tree p) {
   ASSERT (is_tuple (p), "tuple expected");
-  if (get_preference ("interactive questions") == "popup") {
+  if (N(p) == 0) {
+    string ret= object_to_string (call (fun));
+    if (ret != "" && ret != "<unspecified>" && ret != "#<unspecified>")
+      set_message (verbatim (ret), "interactive command");
+  }
+  else if (get_preference ("interactive questions") == "popup" ||
+           N(p) > 1 ||
+	   (is_aux_buffer (get_current_buffer_safe ()) &&
+            !is_rooted_tmfs (get_current_buffer_safe (), "part"))) {
     int i, n= N(p);
     array<string> prompts (n);
     for (i=0; i<n; i++)
       prompts[i]= get_prompt (p, i);
-    command cb= dialogue_command (get_server(), fun, n);
+    command cb= dialogue_command (get_server(), fun, p);
     widget wid= inputs_list_widget (cb, prompts);
     for (i=0; i<n; i++) {
       widget input_wid= get_form_field (wid, i);
@@ -242,17 +258,18 @@ tm_frame_rep::interactive (object fun, scheme_tree p) {
       if (k > 0) set_string_input (input_wid, proposals[0]);
       for (j=0; j<k; j++) add_input_proposal (input_wid, proposals[j]);
     }
-    string title= "Enter data";
-    if (ends (prompts[0], "?")) title= "Question";
+    string title= translate ("Enter data");
+    if (ends (prompts[0], "?")) title= translate ("Question");
     dialogue_start (title, wid);
     send_keyboard_focus (get_form_field (dialogue_wid, 0));
   }
   else {
-    if (get_window () -> get_interactive_mode ()) beep ();
+    if (concrete_window () -> get_interactive_mode ()) beep ();
     else {
       command interactive_cmd=
-	tm_new<interactive_command_rep> (this, get_window (), fun, p);
+        tm_new<interactive_command_rep> (this, concrete_window (), fun, p);
       interactive_cmd ();
     }
   }
 }
+
