@@ -71,6 +71,8 @@ notify_graphics_extents (tree t, point lbot, point rtop) {
 void
 concater_rep::typeset_graphics (tree t, path ip) {
 BEGIN_MAGNIFY
+  env->update_color ();
+  env->update_dash_style_unit ();
   grid gr= as_grid (env->read (GR_GRID));
   array<box> bs;
   gr->set_aspect (env->read (GR_GRID_ASPECT));
@@ -78,9 +80,10 @@ BEGIN_MAGNIFY
                   env->clip_lim1, env->clip_lim2);
   typeset_graphical (bs, t, ip);
 
+  bool crop= env->get_bool (GR_AUTO_CROP);
   point lim1= env->clip_lim1;
   point lim2= env->clip_lim2;
-  if (env->get_bool (GR_AUTO_CROP)) {
+  if (crop) {
     SI x1= MAX_SI, y1= MAX_SI, x2= -MAX_SI, y2= -MAX_SI;
     for (int i=1; i<N(bs); i++) {
       box b= bs[i];
@@ -99,6 +102,14 @@ BEGIN_MAGNIFY
   gr= as_grid (env->read (GR_EDIT_GRID));
   gr->set_aspect (env->read (GR_EDIT_GRID_ASPECT));
   box b= graphics_box (ip, bs, env->fr, gr, lim1, lim2);
+  if (crop && lim1[1] * lim2[1] > 0) {
+    SI y1= env->fr (lim1) [1];
+    SI y2= env->fr (lim2) [1];
+    if (y1 > 0 && y2 > 0)
+      b= move_box (ip, b, 0, -min (y1, y2));
+    else if (y1 < 0 && y2 < 0)
+      b= move_box (ip, b, 0, -max (y1, y2));
+  }
   print (b);
 
   notify_graphics_extents (t, lim1, lim2);
@@ -140,15 +151,42 @@ is_transformation (tree t) {
       is_double (t[1][1]) &&
       is_double (t[2]))
     return true;
+  if (is_tuple (t, "scaling", 2) &&
+      is_double (t[1]) &&
+      is_double (t[2]))
+    return true;
+  if (is_tuple (t, "slanting", 1) &&
+      is_double (t[1]))
+    return true;
+  if (is_tuple (t, "linear", 4) &&
+      is_double (t[1]) &&
+      is_double (t[2]) &&
+      is_double (t[3]) &&
+      is_double (t[4]))
+    return true;
   return false;
 }
 
 frame
 get_transformation (tree t) {
   if (is_tuple (t, "rotation", 1))
-    return rotation_2D (point (0.0, 0.0), as_double (t[1]));
+    return rotation_2D (point (0.0, 0.0), as_double (t[1]) / 57.2957795131);
   if (is_tuple (t, "rotation", 2))
-    return rotation_2D (as_point (t[1]), as_double (t[2]));
+    return rotation_2D (as_point (t[1]), as_double (t[2]) / 57.2957795131);
+  if (is_tuple (t, "scaling", 2))
+    return scaling (point (as_double (t[1]), as_double (t[2])),
+                    point (0.0, 0.0));
+  if (is_tuple (t, "slanting", 1))
+    return slanting (point (0.0, 0.0), as_double (t[1]));
+  if (is_tuple (t, "linear", 4)) {
+    matrix<double> m (0.0, 2, 2);
+    m (0, 0) = as_double (t[1]);
+    m (0, 1) = as_double (t[2]);
+    m (1, 0) = as_double (t[3]);
+    m (1, 1) = as_double (t[4]);
+    return linear_2D (m);
+  }
+  failed_error << "t= " << t << LF;
   FAILED ("transformation expected");
   return frame ();
 }
@@ -194,12 +232,13 @@ BEGIN_MAGNIFY
       else if (halign == "right") x -= b->x2;
       if (valign == "bottom") y -= b->y1;
       else if (valign == "axis") {
-	axis= env->fn->yfrac - b->y1;
-	y -= env->fn->yfrac;
+        axis= env->fn->yfrac - b->y1;
+        y -= env->fn->yfrac;
       }
       else if (valign == "center") y -= ((b->y1 + b->y2) >> 1);
       else if (valign == "top") y -= b->y2;
-      print (text_at_box (ip, b, x, y, axis, env->fn->spc->def));
+      SI pad= env->get_length (TEXT_AT_MARGIN);
+      print (text_at_box (ip, b, x, y, axis, pad));
     }
   }
 END_MAGNIFY
@@ -210,7 +249,7 @@ concater_rep::typeset_math_at (tree t, path ip) {
 BEGIN_MAGNIFY
   if (N(t) != 2) typeset_error (t, ip);
   else {
-    // FIXME: attaching ip to compound ("math", t[0]) is a bit hacky,
+    // FIXME: attaching an ip to compound ("math", t[0]) is a bit hacky,
     // but it seems to work fine for the time being
     box    b     = typeset_as_concat (env, compound ("math", t[0]), ip);
     point  p     = env->fr (env->as_point (env->exec (t[1])));
@@ -218,7 +257,7 @@ BEGIN_MAGNIFY
     string valign= env->text_at_valign;
 
     if (N(p) == 0)
-      typeset_dynamic (tree (ERROR, "bad text-at"), ip);
+      typeset_dynamic (tree (ERROR, "bad math-at"), ip);
     else {
       SI x= (SI) p[0], y= (SI) p[1], axis= (b->h() >> 1);
       if (halign == "left") x -= b->x1;
@@ -226,8 +265,40 @@ BEGIN_MAGNIFY
       else if (halign == "right") x -= b->x2;
       if (valign == "bottom") y -= b->y1;
       else if (valign == "axis") {
-	axis= (env->fn->yx >> 1) - b->y1;
-	y -= (env->fn->yx >> 1);
+        axis= (env->fn->yx >> 1) - b->y1;
+        y -= (env->fn->yx >> 1);
+      }
+      else if (valign == "center") y -= ((b->y1 + b->y2) >> 1);
+      else if (valign == "top") y -= b->y2;
+      print (text_at_box (ip, b, x, y, axis, env->fn->spc->def));
+    }
+  }
+END_MAGNIFY
+}
+
+void
+concater_rep::typeset_document_at (tree t, path ip) {
+BEGIN_MAGNIFY
+  if (N(t) != 2) typeset_error (t, ip);
+  else {
+    // FIXME: attaching ip to compound ("paragraph-box", t[0]) is a bit hacky,
+    // but it seems to work fine for the time being
+    box    b= typeset_as_concat (env, compound ("paragraph-box", t[0]), ip);
+    point  p= env->fr (env->as_point (env->exec (t[1])));
+    string halign= env->text_at_halign;
+    string valign= env->doc_at_valign;
+
+    if (N(p) == 0)
+      typeset_dynamic (tree (ERROR, "bad document-at"), ip);
+    else {
+      SI x= (SI) p[0], y= (SI) p[1], axis= (b->h() >> 1);
+      if (halign == "left") x -= b->x1;
+      else if (halign == "center") x -= ((b->x1 + b->x2) >> 1);
+      else if (halign == "right") x -= b->x2;
+      if (valign == "bottom") y -= b->y1;
+      else if (valign == "axis") {
+        axis= (env->fn->yx >> 1) - b->y1;
+        y -= (env->fn->yx >> 1);
       }
       else if (valign == "center") y -= ((b->y1 + b->y2) >> 1);
       else if (valign == "top") y -= b->y2;
@@ -247,7 +318,8 @@ BEGIN_MAGNIFY
     for (i=0; i<n; i++)
       u[i]= env->exec (t[i]);
     point p= env->fr (env->as_point (u));
-    print (point_box (ip, p, 20*PIXEL, env->pen,
+    pencil pen= env->pen->set_width (env->point_border);
+    print (point_box (ip, p, env->point_size, pen,
                       env->fill_brush, env->point_style));
   }
 END_MAGNIFY
@@ -255,6 +327,7 @@ END_MAGNIFY
 
 static tree
 protect_arrow (edit_env env, tree t) {
+  (void) env;
   return tree (WITH, "arrow-begin", "none", "arrow-end", "none",
                "dash-style", "none", t);
 }
@@ -294,8 +367,8 @@ BEGIN_MAGNIFY
       cip << cip[0];
     }
     curve c= env->fr (poly_segment (a, cip));
-    print (curve_box (ip, c, env->pen,
-                      env->dash_style, env->dash_style_unit,
+    print (curve_box (ip, c, env->line_portion, env->pen,
+                      env->dash_style, env->dash_motif, env->dash_style_unit,
                       env->fill_brush, typeset_line_arrows (ip)));
   }
 END_MAGNIFY
@@ -319,8 +392,8 @@ BEGIN_MAGNIFY
       typeset_line (t, ip, close);
     else {
       curve c= env->fr (arc (a, cip, close));
-      print (curve_box (ip, c, env->pen,
-                        env->dash_style, env->dash_style_unit,
+      print (curve_box (ip, c, env->line_portion, env->pen,
+                        env->dash_style, env->dash_motif, env->dash_style_unit,
                         env->fill_brush, typeset_line_arrows (ip)));
     }
   }
@@ -344,8 +417,8 @@ BEGIN_MAGNIFY
       cip << cip[0];
     }
     curve c= env->fr (N(a)>=3 ? spline (a, cip, close): poly_segment (a, cip));
-    print (curve_box (ip, c, env->pen,
-                      env->dash_style, env->dash_style_unit,
+    print (curve_box (ip, c, env->line_portion, env->pen,
+                      env->dash_style, env->dash_motif, env->dash_style_unit,
                       env->fill_brush, typeset_line_arrows (ip)));
   }
 END_MAGNIFY
@@ -362,6 +435,37 @@ concater_rep::typeset_cspline (tree t, path ip) {
   typeset_spline (t, ip, true);
 }
 
+void
+concater_rep::typeset_bezier (tree t, path ip) {
+BEGIN_MAGNIFY
+  int i, n= N(t);
+  array<point> a(n);
+  for (i=0; i<n; i++)
+    a[i]= env->as_point (env->exec (t[i]));
+  array<path> cip(n);
+  for (i=0; i<n; i++)
+    cip[i]= descend (ip, i);
+  if (N(a) == 0 || N(a[0]) == 0) typeset_error (t, ip);
+  else {
+    if (N(a) == 1) {
+      a << copy (a[0]);
+      cip << cip[0];
+    }
+    curve c;
+    if (N(a) < 3) c= poly_segment (a, cip);
+    else {
+      bool simple= is_func (t, SMOOTH) || is_func (t, CSMOOTH);
+      bool closed= is_func (t, CBEZIER) || is_func (t, CSMOOTH);
+      c= poly_bezier (a, cip, simple, closed);
+    }
+    c= env->fr (c);
+    print (curve_box (ip, c, env->line_portion, env->pen,
+                      env->dash_style, env->dash_motif, env->dash_style_unit,
+                      env->fill_brush, typeset_line_arrows (ip)));
+  }
+END_MAGNIFY
+}
+
 void
 concater_rep::typeset_fill (tree t, path ip) {
   (void) t; (void) ip;
@@ -429,6 +533,70 @@ set_graphical_values (tree t) {
   }
 }
 
+box
+typeset_gr_item (edit_env env, tree t, path ip, bool& ok) {
+  ok= false;
+  if (is_func (t, WITH)) {
+    int i, n= N(t), k= (n-1)>>1; // is k=0 allowed ?
+    if ((n&1) != 1) return empty_box (ip);
+
+    STACK_NEW_ARRAY(vars,string,k);
+    STACK_NEW_ARRAY(oldv,tree,k);
+    STACK_NEW_ARRAY(newv,tree,k);
+    for (i=0; i<k; i++) {
+      tree var_t= env->exec (t[i<<1]);
+      if (is_atomic (var_t)) {
+        string var= var_t->label;
+        vars[i]= var;
+        oldv[i]= env->read (var);
+        newv[i]= env->exec (t[(i<<1)+1]);
+        if (var == PROVISO && newv[i] == "false") {
+          STACK_DELETE_ARRAY(vars);
+          STACK_DELETE_ARRAY(oldv);
+          STACK_DELETE_ARRAY(newv);
+          return empty_box (ip);
+        }
+      }
+      else {
+        STACK_DELETE_ARRAY(vars);
+        STACK_DELETE_ARRAY(oldv);
+        STACK_DELETE_ARRAY(newv);
+        return empty_box (ip);
+      }
+    }
+
+    // for (i=0; i<k; i++) env->monitored_write_update (vars[i], newv[i]);
+    for (i=0; i<k; i++) env->write_update (vars[i], newv[i]);
+    box b= typeset_as_atomic (env, t[n-1], descend (ip, n-1));
+    for (i=k-1; i>=0; i--) env->write_update (vars[i], oldv[i]);
+    STACK_DELETE_ARRAY(vars);
+    STACK_DELETE_ARRAY(oldv);
+    STACK_DELETE_ARRAY(newv);
+    ok= true;
+    return b;
+  }
+  else if ((L(t) == ANIM_STATIC || L(t) == ANIM_DYNAMIC) &&
+	   (N(t) >= 1 && is_func (t[0], MORPH))) {
+    ok= true;
+    for (int i=0; i<N(t[0]); i++)
+      if (is_func (t[0][i], TUPLE, 2) && is_func (t[0][i][1], WITH)) {
+        tree w= t[0][i][1];
+        for (int j=0; j+1<N(w); j++) {
+          if (w[j] == PROVISO && env->exec (w[j+1]) == "false") {
+            ok= false;
+            break;
+          }
+        }
+      }
+    if (!ok) return empty_box (ip);
+    return typeset_as_atomic (env, t, ip);
+  }
+  else {
+    ok= true;
+    return typeset_as_atomic (env, t, ip);
+  }
+}
+
 void
 concater_rep::typeset_graphical (array<box>& bs, tree t, path ip) {
   int i, n= N(t);
@@ -443,16 +611,16 @@ concater_rep::typeset_graphical (array<box>& bs, tree t, path ip) {
           set_graphical_value (u[0], get_graphical_value (u[1]));
         break;
       case IS_INTERSECTION:
-	cout << "Not yet implemented 'is-intersection'\n";
+        cout << "Not yet implemented 'is-intersection'\n";
         break;
       case ON_CURVE:
-	cout << "Not yet implemented 'on-curve'\n";
+        cout << "Not yet implemented 'on-curve'\n";
         break;
       case ON_TEXT_BORDER:
-	cout << "Not yet implemented 'on-text-border'\n";
+        cout << "Not yet implemented 'on-text-border'\n";
         break;
       case ON_GRID:
-	cout << "Not yet implemented 'on-grid'\n";
+        cout << "Not yet implemented 'on-grid'\n";
         break;
       default:
         break;
@@ -460,8 +628,11 @@ concater_rep::typeset_graphical (array<box>& bs, tree t, path ip) {
     }
 
   for (i=0; i<n; i++)
-    if (the_drd->get_type (t[i]) != TYPE_CONSTRAINT && !is_atomic (t[i]))
-      bs << typeset_as_atomic (env, t[i], descend (ip, i));
+    if (the_drd->get_type (t[i]) != TYPE_CONSTRAINT && !is_atomic (t[i])) {
+      bool ok;
+      box b= typeset_gr_item (env, t[i], descend (ip, i), ok);
+      if (ok) bs << b;
+    }
 }
 
 /******************************************************************************
