@@ -17,7 +17,8 @@
 * Cells
 ******************************************************************************/
 
-cell_rep::cell_rep (edit_env env2): var (""), env (env2) {}
+cell_rep::cell_rep (edit_env env2):
+  var (""), env (env2), border_flags (0) {}
 
 void
 cell_rep::typeset (tree fm, tree t, path iq) {
@@ -28,6 +29,7 @@ cell_rep::typeset (tree fm, tree t, path iq) {
     t = t[0];
   }
 
+  cell_local_begin (fm);
   if (is_func (t, SUBTABLE, 1)) {
     lsep= rsep= bsep= tsep= 0;
     T= table (env, 2);
@@ -37,12 +39,13 @@ cell_rep::typeset (tree fm, tree t, path iq) {
     if (hyphen == "n") {
       b= typeset_as_concat (env, t, iq);
       if (vcorrect != "n") {
-	SI y1= b->y1;
-	SI y2= b->y2;
-	if ((vcorrect == "a") || (vcorrect == "b")) y1= min (y1, env->fn->y1);
-	if ((vcorrect == "a") || (vcorrect == "t")) y2= max (y2, env->fn->y2);
-	b= vresize_box (iq, b, y1, y2);
+        SI y1= b->y1;
+        SI y2= b->y2;
+        if ((vcorrect == "a") || (vcorrect == "b")) y1= min (y1, env->fn->y1);
+        if ((vcorrect == "a") || (vcorrect == "t")) y2= max (y2, env->fn->y2);
+        b= vresize_box (iq, b, y1, y2);
       }
+      if (swell > 0) swell_padding ();
     }
     else {
       b= empty_box (iq);
@@ -67,6 +70,7 @@ cell_rep::typeset (tree fm, tree t, path iq) {
       env->local_end (PAGE_MEDIUM, old1);
     }
   }
+  cell_local_end (fm);
 
   if (decoration != "") {
     int i, j, or_row= -1, or_col= -1;
@@ -76,12 +80,12 @@ cell_rep::typeset (tree fm, tree t, path iq) {
       tree dr= dt[i];
       while (is_func (dr, TFORMAT)) dr= dr [N(dr)-1];
       for (j=0; j<N(dr); j++) {
-	tree dc= dr[j];
-	while (is_func (dc, TFORMAT)) dc= dc [N(dc)-1];
-	if (dc == tree (TMARKER)) {
-	  or_row= i;
-	  or_col= j;
-	}
+        tree dc= dr[j];
+        while (is_func (dc, TFORMAT)) dc= dc [N(dc)-1];
+        if (dc == tree (TMARKER)) {
+          or_row= i;
+          or_col= j;
+        }
       }
     }
 
@@ -92,6 +96,36 @@ cell_rep::typeset (tree fm, tree t, path iq) {
   }
 }
 
+void
+cell_rep::cell_local_begin (tree fm) {
+  int i, l= N(fm);
+  for (i=0; i<l; i++) {
+    tree with= fm[i];
+    if (is_func (with, CWITH, 2) &&
+        is_atomic (with[0]) &&
+        !starts (with[0]->label, "cell-")) {
+      tree old= env->local_begin (with[0]->label, with[1]);
+      string v= with[0]->label * "-" * as_string (i);
+      var (v)= old;
+    }
+  }
+}
+
+void
+cell_rep::cell_local_end (tree fm) {
+  int i, l= N(fm);
+  for (i=l-1; i>=0; i--) {
+    tree with= fm[i];
+    if (is_func (with, CWITH, 2) &&
+        is_atomic (with[0]) &&
+        !starts (with[0]->label, "cell-")) {
+      string v= with[0]->label * "-" * as_string (i);
+      tree old= var [v];
+      env->local_end (with[0]->label, old);
+    }
+  }
+}
+
 /******************************************************************************
 * Formatting routines
 ******************************************************************************/
@@ -104,20 +138,20 @@ extract_format (tree fm, tree* r, int n) {
   for (i=0; i<N(fm); i++)
     if (is_func (fm[i], CWITH))
       if ((N(fm[i]) >= 2) &&
-	  (is_int (fm[i][0])) &&
-	  (is_int (fm[i][1])))
-	{
-	  int k;
-	  int k1= as_int (fm[i][0]);
-	  int k2= as_int (fm[i][1]);
-	  tree u= fm[i] (2, N (fm[i]));
-	  if (k1>=0) k1--; else k1+=n;
-	  if (k2> 0) k2--; else k2+=n;
-	  if ((k1 >= n) || (k2 < 0)) continue;
-	  k1= max (k1, 0);
-	  k2= min (k2, n-1);
-	  for (k=k1; k<=k2; k++) r[k] << u;
-	}
+          (is_int (fm[i][0])) &&
+          (is_int (fm[i][1])))
+        {
+          int k;
+          int k1= as_int (fm[i][0]);
+          int k2= as_int (fm[i][1]);
+          tree u= fm[i] (2, N (fm[i]));
+          if (k1>=0) k1--; else k1+=n;
+          if (k2> 0) k2--; else k2+=n;
+          if ((k1 >= n) || (k2 < 0)) continue;
+          k1= max (k1, 0);
+          k2= min (k2, n-1);
+          for (k=k1; k<=k2; k++) r[k] << u;
+        }
 }
 
 void
@@ -202,6 +236,9 @@ cell_rep::format_cell (tree fm) {
   if (var->contains (CELL_COL_SPAN))
     col_span= as_int (env->exec (var[CELL_COL_SPAN]));
   else col_span= 1;
+  if (var->contains (CELL_SWELL))
+    swell= env->as_length (env->exec (var[CELL_SWELL])) >> 1;
+  else swell= 0;
 }
 
 void
@@ -218,11 +255,11 @@ void
 cell_rep::compute_width (SI& mw, SI& lw, SI& rw, bool large) {
   //cout << "  large= " << large << "\n";
   char align_c= '\0'; if (N (halign) != 0) align_c= halign[0];
-  bool lr_flag= (align_c >= 'A' && align_c <= 'Z');
+  bool lr_flag= is_upcase (align_c);
   if (!is_nil (T)) {
     if (N (T->halign) > 0) align_c= T->halign[0];
     else align_c= '\0';
-    lr_flag= (align_c >= 'A' && align_c <= 'Z');
+    lr_flag= is_upcase (align_c);
     T->compute_width (mw, lw, rw);
     if (lr_flag) {
       lw += lborder;
@@ -246,15 +283,15 @@ cell_rep::compute_width (SI& mw, SI& lw, SI& rw, bool large) {
     lw= rw= mw= 0;
     if (lr_flag) {
       if (N (halign) == 1) {
-	lw= -b->x1+ lsep + lborder;
-	rw=  b->x2+ rsep + rborder;
-	mw=  lw + rw;
+        lw= -b->x1+ lsep + lborder;
+        rw=  b->x2+ rsep + rborder;
+        mw=  lw + rw;
       }
       else {
-	SI offset= b->get_leaf_offset (halign (1, N (halign)));
-	lw= offset+ lsep+ lborder;
-	rw= b->w()- offset+ rsep+ rborder;
-	mw= lw + rw;
+        SI offset= b->get_leaf_offset (halign (1, N (halign)));
+        lw= offset+ lsep+ lborder;
+        rw= b->w()- offset+ rsep+ rborder;
+        mw= lw + rw;
       }
     }
     else mw= b->w() + lsep + rsep + lborder + rborder;
@@ -276,7 +313,7 @@ cell_rep::compute_height (SI& mh, SI& bh, SI& th) {
   char align_c= '\0'; if (N (valign) != 0) align_c= valign[0];
   if (is_nil (T)) {
     bh= th= mh= 0;
-    if (align_c >= 'A' && align_c <= 'Z') {
+    if (is_upcase (align_c)) {
       bh= -b->y1+ bsep + bborder;
       th=  b->y2+ tsep + tborder;
       mh=  bh + th;
@@ -286,7 +323,7 @@ cell_rep::compute_height (SI& mh, SI& bh, SI& th) {
   else {
     if (N (T->valign) != 0) align_c= T->valign[0];
     T->compute_height (mh, bh, th);
-    if (align_c >= 'A' && align_c <= 'Z') {
+    if (is_upcase (align_c)) {
       bh += bborder;
       th += tborder;
     }
@@ -344,6 +381,28 @@ cell_rep::position_vertically (SI offset, SI mh, SI bh, SI th) {
 * Producing the final cell box
 ******************************************************************************/
 
+void
+cell_rep::swell_padding () {
+  if (row_span > 1) return;
+  SI swt= env->get_length (MATH_TOP_SWELL_START);
+  SI swb= env->get_length (MATH_BOT_SWELL_START);
+  if (b->y2 > swt && (border_flags & 1) == 0) {
+    SI swT= env->get_length (MATH_TOP_SWELL_END);
+    double exceed= b->y2 - swt;
+    double unit  = max (swT - swt, 1);
+    double ratio = min (exceed / unit, 1.0);
+    tsep += (SI) (ratio * swell);
+  }
+  if (b->y1 < swb && (border_flags & 2) == 0) {
+    SI swB= env->get_length (MATH_BOT_SWELL_END);
+    double exceed= swb - b->y1;
+    double unit  = max (swb - swB, 1);
+    double ratio = min (exceed / unit, 1.0);
+    bsep += (SI) (ratio * swell);
+  }
+  swell= 0;
+}
+
 void
 cell_rep::finish_horizontal () {
   SI  w= width- lsep- lborder- rsep- rborder;
@@ -351,6 +410,7 @@ cell_rep::finish_horizontal () {
   SI  d= ((vcorrect == "b") || (vcorrect == "a"))? -env->fn->y1: 0;
   SI  h= ((vcorrect == "t") || (vcorrect == "a"))?  env->fn->y2: 0;
   b= (box) lz->produce (LAZY_BOX, make_format_cell (w, v, d, h));
+  if (swell > 0) swell_padding ();
 }
 
 void
@@ -361,6 +421,6 @@ cell_rep::finish () {
   }
 
   b= cell_box (ip, b, xoff, yoff, 0, 0, x2-x1, y2-y1,
-	       lborder, rborder, bborder, tborder,
+               lborder, rborder, bborder, tborder,
                env->pen->get_brush (), brush (bg, env->alpha));
 }
